"""."""

import warnings
from abc import abstractmethod
from collections.abc import Callable, Iterable
from typing import Any, cast, Self

import equinox as eqx
import numpy as np
import pandas as pd

from ..base import AbstractConfig
from ..coding_scheme import (
    CodeMap,
    CodingScheme,
    CodingSchemesManager,
    CodingSchemeWithUOM,
    FrozenDict11,
    NumericScheme,
    ReducedCodeMapN1,
)
from ..dataset import (
    AdmissionIntervalEventsTableColumns,
    AdmissionIntervalRatesTableColumns,
    AdmissionSummaryTableColumns,
    AdmissionTimeSeriesTableColumns,
    COLUMN,
    Dataset,
    DatasetConfig,
    DatasetSchemeConfig,
    DatasetTables,
    MultivariateTimeSeriesTableMeta,
    SECONDS_TO_HOURS_SCALER,
    StaticTableColumns,
    TableColumns,
)
from ..example_schemes.icd_ccs_integration import setup_standard_icd_ccs
from ..example_schemes.mixed_icd import MultiVersionScheme
from ..utils import resources_path


warnings.filterwarnings("error", category=RuntimeWarning, message=r"overflow encountered in cast")


class TableResource(AbstractConfig):
    columns: TableColumns

    def __init__(self, columns: TableColumns):
        self.columns = columns

    @staticmethod
    def _coerce_columns_to_str(df: pd.DataFrame, columns: Iterable[str]) -> pd.DataFrame:
        is_str = pd.api.types.is_string_dtype
        # coerce to integers then fix as strings.
        int_dtypes = {k: int for k in columns if k in df.columns and not is_str(df.dtypes[k])}
        str_dtypes = {k: str for k in columns if k in df.columns and not is_str(df.dtypes[k])}
        return df.astype(int_dtypes).astype(str_dtypes)

    def _coerce_id_to_str(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Some of the integer ids in the database when downloaded are stored as floats.
        A fix is to coerce them to integers then fix as strings.
        """
        return self._coerce_columns_to_str(df, self.columns.id_cols)

    def _coerce_code_to_str(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Some of the integer codes in the database when downloaded are stored as floats or integers.
        A fix is to coerce them to integers then fix as strings.
        """
        return self._coerce_columns_to_str(df, self.columns.code_cols)

    @staticmethod
    def _filter_null_codes(df: pd.DataFrame) -> pd.DataFrame:
        """
        Some code cells are empty (e.g., ROW_ID=7969 in MIMIC-III v1.3 DIAGNOSES_ICD.csv.gz). Filter these rows.
        """
        not_null = df[COLUMN.code].notnull() & (df[COLUMN.code] != "")
        return df.loc[not_null]

    @property
    def pipeline(self) -> tuple[Callable[[pd.DataFrame], pd.DataFrame], ...]:
        return (self._coerce_id_to_str,)

    @staticmethod
    def apply_pipeline(pipeline: tuple[Callable[[pd.DataFrame], pd.DataFrame], ...], table):
        for f in pipeline:
            table = f(table)
        return table

    @abstractmethod
    def load_standard_columns_table(self, data_connection: Any, **kwargs) -> pd.DataFrame:
        raise NotImplementedError()

    def __call__(self, data_connection: Any, **kwargs) -> pd.DataFrame:
        return self.apply_pipeline(self.pipeline, self.load_standard_columns_table(data_connection, **kwargs))


CodedColumns = (
    AdmissionSummaryTableColumns
    | AdmissionTimeSeriesTableColumns
    | AdmissionIntervalEventsTableColumns
    | AdmissionIntervalRatesTableColumns
)


class CodedTableResource(TableResource):
    def __check_init__(self):
        assert all(c in self.columns for c in (COLUMN.code, COLUMN.description))

    @property
    def pipeline(self) -> tuple[Callable[[pd.DataFrame], pd.DataFrame], ...]:
        return self._coerce_id_to_str, self._filter_null_codes, self._coerce_code_to_str

    def space(self, data_connection: Any) -> pd.DataFrame:
        return self.apply_pipeline(self.pipeline, self.load_space_table(data_connection))

    @abstractmethod
    def load_space_table(self, data_connection: Any):
        raise NotImplementedError("This method should be implemented in subclasses.")


class StaticTableResource(TableResource):
    columns: StaticTableColumns

    def __init__(self):
        super().__init__(StaticTableColumns())

    @staticmethod
    def substitute_null_function(column: str, replace: str) -> Callable[[pd.DataFrame], pd.DataFrame]:
        def _apply(df: pd.DataFrame) -> pd.DataFrame:
            return df.assign(**{column: df[column].fillna(value=replace).astype(str)})

        return _apply

    @abstractmethod
    def load_gender_space_table(self, data_connection: Any) -> pd.DataFrame:
        raise NotImplementedError()

    @abstractmethod
    def load_ethnicity_space_table(self, data_connection: Any) -> pd.DataFrame:
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def derive_deshifted_date_of_birth(cls, patients: pd.DataFrame, **kwargs) -> pd.DataFrame:
        # Different procedures to implement for MIMIC-III and MIMIC-IV
        raise NotImplementedError()

    @classmethod
    def _add_deshifted_date_of_birth(cls, admissions: pd.DataFrame) -> Callable[[pd.DataFrame], pd.DataFrame]:
        def _add(df: pd.DataFrame) -> pd.DataFrame:
            df[COLUMN.date_of_birth] = cls.derive_deshifted_date_of_birth(df, admissions=admissions)
            return df

        return _add

    @property
    def pipeline(self) -> None:
        return None

    def gender_space(self, date_source: Any) -> pd.DataFrame:
        p = self.substitute_null_function(str(COLUMN.gender), "MISSING_GENDER")
        return p(self.load_gender_space_table(date_source))

    def ethnicity_space(self, data_connection: Any) -> pd.DataFrame:
        p = self.substitute_null_function(str(COLUMN.race), "MISSING_ETHNICITY")
        return p(self.load_ethnicity_space_table(data_connection))

    def __call__(self, data_connection: Any, **kwargs) -> pd.DataFrame:
        assert "admissions" in kwargs, "Pass the processed admissions table."
        admissions = kwargs.pop("admissions")
        pipeline = (
            self._coerce_id_to_str,
            self._add_deshifted_date_of_birth(admissions=admissions),
            self.substitute_null_function(str(COLUMN.race), "MISSING_ETHNICITY"),
            self.substitute_null_function(str(COLUMN.gender), "MISSING_GENDER"),
        )
        return self.apply_pipeline(pipeline, self.load_standard_columns_table(data_connection, **kwargs))


class StaticTableResource_MIMICIV(StaticTableResource):
    @classmethod
    def derive_deshifted_date_of_birth(cls, patients: pd.DataFrame, **kwargs) -> pd.Series:
        return pd.Series(
            list(
                map(
                    lambda dt, age: dt + pd.DateOffset(years=-age),
                    pd.to_datetime(patients[COLUMN.anchor_year], format="%Y").dt.normalize(),
                    patients[COLUMN.anchor_age].astype(int),
                )
            ),
            index=patients.index,
        )


class StaticTableResource_MIMICIII(StaticTableResource):
    @classmethod
    def derive_deshifted_date_of_birth(cls, patients: pd.DataFrame, **kwargs) -> pd.Series:
        """
        Important comment from MIMIC-III documentation at \
            https://mimic.mit.edu/docs/iii/tables/patients/
        > DOB is the date of birth of the given patient. Patients who are \
            older than 89 years old at any time in the database have had their\
            date of birth shifted to obscure their age and comply with HIPAA.\
            The shift process was as follows: the patient’s age at their \
            first admission was determined. The date of birth was then set to\
            exactly 300 years before their first admission.

        # TODO: check https://mimic.mit.edu/docs/iii/about/time/
        """
        assert "admissions" in kwargs, "Should pass the admissions table as keyword argument."
        admissions = kwargs.pop("admissions").loc[:, [COLUMN.subject_id, COLUMN.start_time, COLUMN.end_time]]
        admissions[COLUMN.start_time] = pd.to_datetime(admissions[COLUMN.start_time])
        admissions[COLUMN.end_time] = pd.to_datetime(admissions[COLUMN.end_time])
        dob = pd.to_datetime(patients[COLUMN.date_of_birth])
        last_disch_date = admissions.groupby(COLUMN.subject_id)[COLUMN.end_time].max()
        first_admit_date = admissions.groupby(COLUMN.subject_id)[COLUMN.start_time].min()
        last_disch_date = last_disch_date.loc[patients[COLUMN.subject_id]]
        first_admit_date = first_admit_date.loc[patients[COLUMN.subject_id]]
        uncertainty = (last_disch_date.dt.year.values - first_admit_date.dt.year.values) // 2
        age_before_shift = uncertainty + 89
        deshifted = np.array(list(map(lambda dt, s: dt + pd.DateOffset(-s), first_admit_date, age_before_shift)))
        adjusted_dob = np.where((last_disch_date.dt.year.values - dob.dt.year.values) > 150, deshifted, dob.values)
        return pd.Series(pd.to_datetime(adjusted_dob), index=dob.index).dt.normalize()


class MixedVersionICDSummaryTableColumns(TableColumns):
    admission_id: str = str(COLUMN.admission_id)
    code: str = str(COLUMN.code)
    version: str = str(COLUMN.version)
    description: str = str(COLUMN.description)


class MixedICDTableResource(CodedTableResource):
    columns: MixedVersionICDSummaryTableColumns

    def __init__(self):
        super().__init__(MixedVersionICDSummaryTableColumns())

    def setup_schemes(
        self,
        manager: CodingSchemesManager,
        name: str,
        component_schemes: dict[str, str],
        infer_maps: tuple[str, ...],
        target_name: str | None,
        mapping: pd.DataFrame | None,
        selection: pd.DataFrame | None,
        *,
        c_code: str,
        c_version: str,
        c_target_code: str,
        c_target_desc: str,
    ) -> CodingSchemesManager:
        scheme = self.register_scheme(
            name=name,
            selection=selection,
            component_schemes={k: manager.scheme[v] for k, v in component_schemes.items()},
        )
        manager = manager.add_scheme(scheme)
        if target_name is not None and mapping is not None:
            manager = scheme.register_map(
                manager=manager,
                target_name=target_name,
                mapping=mapping,
                c_code=c_code,
                c_version=c_version,
                c_target_code=c_target_code,
                c_target_desc=c_target_desc,
            )
        for target in infer_maps:
            manager = scheme.register_infer_map(manager, target)
        return manager

    @staticmethod
    def register_scheme(
        name: str, component_schemes: dict[str, MultiVersionScheme], selection: pd.DataFrame | None
    ) -> MultiVersionScheme:
        return MultiVersionScheme.from_selection(name, selection, component_schemes=component_schemes)

    @staticmethod
    @abstractmethod
    def _add_version_column_if_not_exists(df: pd.DataFrame) -> pd.DataFrame:
        # This is specifically added for MIMIC-III, pure ICD-9 codes.
        raise NotImplementedError("Override this method in subclass")

    @classmethod
    def _coerce_version_to_str(cls, df: pd.DataFrame) -> pd.DataFrame:
        """
        Some of the integer codes in the database when downloaded are stored as floats or integers.
        A fix is to coerce them to integers then fix as strings.
        """
        return cls._coerce_columns_to_str(df, (str(COLUMN.version),))

    @staticmethod
    def _strip_icd_codes(df: pd.DataFrame) -> pd.DataFrame:
        df[COLUMN.code] = df[COLUMN.code].str.strip()
        return df

    @staticmethod
    def _mixed_code_format(
        mixed_scheme_name: str, schemes_manager: CodingSchemesManager
    ) -> Callable[[pd.DataFrame], pd.DataFrame]:
        scheme = cast(MultiVersionScheme, schemes_manager.scheme[mixed_scheme_name])

        def _transform(df: pd.DataFrame) -> pd.DataFrame:
            return scheme.mixed_code_format_table(schemes_manager, df)

        return _transform

    @property
    def pipeline(self) -> None:
        return None

    def space(self, data_connection: Any) -> pd.DataFrame:
        pipeline = (
            self._coerce_code_to_str,
            self._filter_null_codes,
            self._strip_icd_codes,
            self._add_version_column_if_not_exists,
            self._coerce_version_to_str,
        )
        return self.apply_pipeline(pipeline, self.load_space_table(data_connection))

    def _custom_pipeline(
        self, mixed_scheme_name: str, schemes_manager: CodingSchemesManager
    ) -> tuple[Callable[[pd.DataFrame], pd.DataFrame], ...]:
        return (
            self._coerce_id_to_str,
            self._filter_null_codes,
            self._coerce_code_to_str,
            self._strip_icd_codes,
            self._add_version_column_if_not_exists,
            self._coerce_version_to_str,
            self._mixed_code_format(mixed_scheme_name, schemes_manager),
        )

    def __call__(self, data_connection: Any, **kwargs) -> pd.DataFrame:
        mixed_scheme_name = kwargs.pop("mixed_scheme_name")
        schemes_manager = kwargs.pop("schemes_manager")
        pipeline = self._custom_pipeline(mixed_scheme_name, schemes_manager)
        return self.apply_pipeline(pipeline, self.load_standard_columns_table(data_connection, **kwargs))


class MixedICDTableResource_MIMICIII(MixedICDTableResource):
    @staticmethod
    def _add_version_column_if_not_exists(df: pd.DataFrame) -> pd.DataFrame:
        assert str(COLUMN.version) not in df
        df = df.iloc[:]
        df[COLUMN.version] = "9"
        return df


class MixedICDTableResource_MIMICIV(MixedICDTableResource):
    @staticmethod
    def _add_version_column_if_not_exists(df: pd.DataFrame) -> pd.DataFrame:
        assert str(COLUMN.version) in df.columns
        return df


class MultivariateTimeSeriesTableResource(CodedTableResource):
    config: MultivariateTimeSeriesTableMeta
    columns: AdmissionTimeSeriesTableColumns

    def __init__(self, config: MultivariateTimeSeriesTableMeta):
        super().__init__(AdmissionTimeSeriesTableColumns())
        self.config = config

    @staticmethod
    def _validate_columns(table: pd.DataFrame, attributes: tuple[str, ...]) -> pd.DataFrame:
        # Perform any necessary validations on the table.
        assert all(attr not in COLUMN for attr in attributes), f"Attributes {attributes} must not be in COLUMN enum."
        assert len(set(attributes)) == len(attributes), f"Duplicate attributes {attributes} found."
        assert all(c in COLUMN for c in set(table.columns) - set(attributes)), (
            f"Some columns {set(table.columns) - set(attributes)} are not in COLUMN enum."
        )
        return table

    @staticmethod
    def _melt_attributes(table: pd.DataFrame) -> pd.DataFrame:
        melted_obs_df = table.melt(
            id_vars=[str(COLUMN.admission_id), str(COLUMN.time)],
            var_name=str(COLUMN.code),
            value_name=str(COLUMN.measurement),
        )
        return melted_obs_df[melted_obs_df[COLUMN.measurement].notnull()]

    @staticmethod
    def _coerce_value_to_real(df: pd.DataFrame) -> pd.DataFrame:
        """
        Some values in the measurement column were found befoer to be stored and loaded as strings.
        """
        return df.astype({str(COLUMN.measurement): float})

    def _rename_attributes(self, df: pd.DataFrame) -> pd.DataFrame:
        df[COLUMN.code] = pd.Series([self.config.name] * len(df)) + "." + df[COLUMN.code]
        return df

    @property
    def pipeline(self) -> tuple[Callable[[pd.DataFrame], pd.DataFrame], ...]:
        return (
            lambda df: self._validate_columns(df, self.config.attributes),
            self._coerce_id_to_str,
            self._melt_attributes,
            self._rename_attributes,
            self._coerce_code_to_str,
            self._coerce_value_to_real,
        )

    def load_space_table(self, data_sourece: None = None) -> pd.DataFrame:
        space = pd.DataFrame(
            [(self.config.name, a, self.config.type_hint[i]) for i, a in enumerate(self.config.attributes)],
            columns=["group", "attribute", "type_hint"],
        )
        space["code"] = space["group"] + "." + space["attribute"]
        return space.sort_values("code")

    def space(self, data_connection: None = None):
        # to skip applying pipeline
        return self.load_space_table(None)


class GroupedMultivariateTimeSeriesTableResource(CodedTableResource):
    groups: tuple[MultivariateTimeSeriesTableResource, ...]

    def __init__(self, groups: tuple[MultivariateTimeSeriesTableResource, ...]):
        super().__init__(AdmissionTimeSeriesTableColumns())
        self.groups = groups

    @staticmethod
    def _time_stat(code_table: pd.DataFrame) -> pd.Series:
        c_admission_id = str(COLUMN.admission_id)
        c_time = str(COLUMN.time)
        timestamps = code_table[[c_admission_id, c_time]].sort_values([c_admission_id, c_time])
        time_deltas = (timestamps[c_time].diff().dt.total_seconds() * SECONDS_TO_HOURS_SCALER).iloc[1:]
        in_admission = pd.Series(timestamps[c_admission_id] == timestamps[c_admission_id].shift()).iloc[1:]
        time_deltas_stats = time_deltas[in_admission].describe()
        return time_deltas_stats.rename(index={k: f"time_delta_{k}" for k in time_deltas_stats.index})

    @staticmethod
    def _stats(code: str, code_table: pd.DataFrame) -> pd.DataFrame:
        values = code_table[COLUMN.measurement]
        stats = values.describe()
        stats["nunique"] = values.nunique()
        time_stats = GroupedMultivariateTimeSeriesTableResource._time_stat(code_table)
        stats = pd.concat([stats, time_stats])
        stats = stats.to_frame().T
        stats[COLUMN.code] = code
        return stats.set_index(str(COLUMN.code))

    def load_standard_columns_table(self, data_connection: Any, **kwargs) -> pd.DataFrame:
        return pd.concat([g(data_connection, **kwargs) for g in self.groups], axis=0)

    @property
    def pipeline(self) -> tuple[Callable[[pd.DataFrame], pd.DataFrame], ...]:
        return tuple([lambda df: df.reset_index(drop=True)])

    def stats(self, data_connection: Any) -> pd.DataFrame:
        dfs = []
        for g in self.groups:
            for code, code_table in g(data_connection).groupby(str(COLUMN.code)):
                dfs.append(self._stats(code, code_table))
        return pd.concat(dfs, axis=0)

    def load_space_table(self, data_sourece: None = None) -> pd.DataFrame:
        return pd.concat([c.space(None) for c in self.groups]).sort_values(["code"])

    def space(self, data_connection: None = None) -> pd.DataFrame:
        return self.load_space_table(None)

    @staticmethod
    def register_scheme(
        name: str, space_table: pd.DataFrame, attributes_selection: pd.DataFrame | None
    ) -> CodingSchemesManager:
        if attributes_selection is None:
            attributes_selection = space_table
        else:
            if "type_hint" not in attributes_selection.columns:
                attributes_selection = pd.merge(
                    attributes_selection,
                    space_table,
                    left_on=["group", "attribute"],
                    right_on=["group", "attribute"],
                    suffixes=(None, "_y"),
                    how="inner",
                )

        # format codes to be of the form 'group.attribute'
        df = attributes_selection.astype({"attribute": str, "type_hint": str, "group": str})
        codes = tuple(df["group"] + "." + df["attribute"].tolist())
        desc = FrozenDict11(dict(zip(codes, codes)))
        group = FrozenDict11(dict(zip(codes, df.index.astype(str).values)))
        type_hint = FrozenDict11(dict(zip(codes, df["type_hint"].tolist())))
        scheme = NumericScheme(name=name, codes=tuple(sorted(codes)), group=group, desc=desc, type_hint=type_hint)  # noqa
        return CodingSchemesManager().add_scheme(scheme)


class DatasetSchemeMapsFileNames(AbstractConfig):
    gender: str | None = "gender.csv"
    ethnicity: str | None = "ethnicity.csv"
    icu_inputs: str | None = "icu_inputs.csv"
    icu_procedures: str | None = "icu_procedures.csv"
    hosp_procedures: str | None = "hosp_procedures.csv"
    dx_discharge: str | None = "dx_discharge.csv"


class ExternalMapResources(AbstractConfig):
    filenames: DatasetSchemeMapsFileNames
    resources_dir: str

    def __init__(self, resources_dir: str, filenames: DatasetSchemeMapsFileNames = DatasetSchemeMapsFileNames()):
        self.filenames = filenames
        self.resources_dir = resources_dir

    def map_file(self, filename: str) -> pd.DataFrame | None:
        try:
            return pd.read_csv(resources_path(self.resources_dir, filename)).astype(str)
        except FileNotFoundError:
            return None

    @property
    def dx_discharge(self) -> pd.DataFrame | None:
        return self.map_file(self.filenames.dx_discharge)

    @property
    def gender(self) -> pd.DataFrame | None:
        return self.map_file(self.filenames.gender)

    @property
    def ethnicity(self) -> pd.DataFrame | None:
        return self.map_file(self.filenames.ethnicity)

    @property
    def icu_inputs(self) -> pd.DataFrame | None:
        return self.map_file(self.filenames.icu_inputs)

    @property
    def icu_procedures(self) -> pd.DataFrame | None:
        return self.map_file(self.filenames.icu_procedures)

    @property
    def hosp_procedures(self) -> pd.DataFrame | None:
        return self.map_file(self.filenames.hosp_procedures)


class DatasetSchemeSelectionFiles(AbstractConfig):
    gender: str | None = "gender.csv"
    ethnicity: str | None = "ethnicity.csv"
    icu_inputs: str | None = "icu_inputs.csv"
    icu_procedures: str | None = "icu_procedures.csv"
    hosp_procedures: str | None = "hosp_procedures.csv"
    obs: str | None = "obs.csv"
    dx_discharge: str | None = "dx_discharge.csv"


class ExternalSelectionResources(AbstractConfig):
    filenames: DatasetSchemeSelectionFiles
    resources_dir: str

    def __init__(self, resources_dir: str, filenames: DatasetSchemeSelectionFiles = DatasetSchemeSelectionFiles()):
        self.filenames = filenames
        self.resources_dir = resources_dir

    def selection_file(self, path: str) -> pd.DataFrame | None:
        try:
            return pd.read_csv(resources_path(self.resources_dir, path)).astype(str)
        except FileNotFoundError:
            return None

    @property
    def gender(self) -> pd.DataFrame | None:
        return self.selection_file(self.filenames.gender)

    @property
    def ethnicity(self) -> pd.DataFrame | None:
        return self.selection_file(self.filenames.ethnicity)

    @property
    def icu_inputs(self) -> pd.DataFrame | None:
        return self.selection_file(self.filenames.icu_inputs)

    @property
    def icu_procedures(self) -> pd.DataFrame | None:
        return self.selection_file(self.filenames.icu_procedures)

    @property
    def hosp_procedures(self) -> pd.DataFrame | None:
        return self.selection_file(self.filenames.hosp_procedures)

    @property
    def obs(self) -> pd.DataFrame | None:
        df = self.selection_file(self.filenames.obs)
        return df.set_index("group", drop=True).sort_values("attribute").sort_index()

    @property
    def dx_discharge(self) -> pd.DataFrame | None:
        return self.selection_file(self.filenames.dx_discharge)


class MIMICDatasetSchemeSuffixes(AbstractConfig):
    gender: str | None
    ethnicity: str | None
    dx_discharge: str | None
    obs: str | None
    icu_inputs: str | None
    icu_procedures: str | None
    hosp_procedures: str | None

    def __init__(
        self,
        gender: str | None = None,
        ethnicity: str | None = None,
        dx_discharge: str | None = None,
        obs: str | None = None,
        icu_inputs: str | None = None,
        icu_procedures: str | None = None,
        hosp_procedures: str | None = None,
    ):
        self.gender = gender
        self.ethnicity = ethnicity
        self.dx_discharge = dx_discharge
        self.obs = obs
        self.icu_inputs = icu_inputs
        self.icu_procedures = icu_procedures
        self.hosp_procedures = hosp_procedures


class ScopedSchemeNames(AbstractConfig):
    suffixes: MIMICDatasetSchemeSuffixes
    name_separator: str
    name_prefix: str
    global_infix: tuple[str, ...] = ()

    def __init__(
        self,
        name_separator: str = ".",
        name_prefix: str = "",
        suffixes: MIMICDatasetSchemeSuffixes = MIMICDatasetSchemeSuffixes(),
        global_infix: tuple[str, ...] = (),
    ):
        self.suffixes = suffixes
        self.name_separator = name_separator
        self.name_prefix = name_prefix
        self.global_infix = global_infix

    def _scheme_name(self, k: str) -> str | None:
        suffix = getattr(self.suffixes, k)
        if suffix is None:
            return None
        return self.name_separator.join((self.name_prefix,) + self.global_infix + (suffix,))

    @property
    def gender(self) -> str | None:
        return self._scheme_name("gender")

    @property
    def ethnicity(self) -> str | None:
        return self._scheme_name("ethnicity")

    @property
    def icu_inputs(self) -> str | None:
        return self._scheme_name("icu_inputs")

    @property
    def icu_procedures(self) -> str | None:
        return self._scheme_name("icu_procedures")

    @property
    def hosp_procedures(self) -> str | None:
        return self._scheme_name("hosp_procedures")

    @property
    def dx_discharge(self) -> str | None:
        return self._scheme_name("dx_discharge")

    @property
    def obs(self) -> str | None:
        return self._scheme_name("obs")

    @property
    def mapped(self) -> Self:
        return eqx.tree_at(lambda x: x.global_infix, self, self.global_infix + ("mapped",))

    def column_name(self, key: str) -> str:
        return self.name_separator.join(self.global_infix + (key,))


class MIMICDatasetAuxiliaryResources(AbstractConfig):
    scoped_names: ScopedSchemeNames
    maps: ExternalMapResources
    selections: ExternalSelectionResources
    icu_inputs_uom_normalization: str | None
    icu_inputs_aggregation_column: str | None

    def __init__(
        self,
        maps: ExternalMapResources,
        selections: ExternalSelectionResources,
        scoped_names: ScopedSchemeNames = ScopedSchemeNames(),
        icu_inputs_uom_normalization: str | None = None,
        icu_inputs_aggregation_column: str | None = None,
    ):
        self.scoped_names = scoped_names
        self.maps = maps
        self.selections = selections
        self.icu_inputs_uom_normalization = icu_inputs_uom_normalization
        self.icu_inputs_aggregation_column = icu_inputs_aggregation_column

    @classmethod
    def make_resources(
        cls,
        name_prefix: str,
        resources_root: str,
        suffixes: MIMICDatasetSchemeSuffixes = MIMICDatasetSchemeSuffixes(),
        name_separator: str = ".",
        selection_subdir: str = "selection",
        map_subdir: str = "map",
        map_files: DatasetSchemeMapsFileNames = DatasetSchemeMapsFileNames(),
        selection_files: DatasetSchemeSelectionFiles = DatasetSchemeSelectionFiles(),
        icu_inputs_uom_normalization: tuple[str, ...] | None = None,
        icu_inputs_aggregation_column: str | None = None,
    ):
        scoped_names = ScopedSchemeNames(suffixes=suffixes, name_separator=name_separator, name_prefix=name_prefix)
        maps = ExternalMapResources(resources_path(resources_root, map_subdir), filenames=map_files)
        selections = ExternalSelectionResources(
            resources_path(resources_root, selection_subdir), filenames=selection_files
        )
        if icu_inputs_uom_normalization is not None:
            icu_inputs_uom_normalization = resources_path(resources_root, *icu_inputs_uom_normalization)

        icu_inputs_aggregation_column = icu_inputs_aggregation_column
        return cls(
            scoped_names=scoped_names,
            maps=maps,
            selections=selections,
            icu_inputs_uom_normalization=icu_inputs_uom_normalization,
            icu_inputs_aggregation_column=icu_inputs_aggregation_column,
        )

    @property
    def icu_inputs_uom_normalization_table(self) -> pd.DataFrame:
        assert self.icu_inputs_uom_normalization is not None, "icu_inputs_uom_normalization is not set"
        return pd.read_csv(self.icu_inputs_uom_normalization).astype(str)


class DatasetTablesResources(AbstractConfig):
    static: StaticTableResource
    admissions: TableResource
    dx_discharge: MixedICDTableResource
    obs: GroupedMultivariateTimeSeriesTableResource
    hosp_procedures: MixedICDTableResource
    icu_procedures: CodedTableResource
    icu_inputs: CodedTableResource


class MIMICSchemeResources(AbstractConfig):
    tables: DatasetTablesResources
    scheme: DatasetSchemeConfig
    aux: MIMICDatasetAuxiliaryResources

    def __init__(
        self, tables: DatasetTablesResources, scheme: DatasetSchemeConfig, aux: MIMICDatasetAuxiliaryResources
    ):
        self.scheme = scheme
        self.tables = tables
        self.aux = aux

    def _make_demographic_scheme(
        self,
        name: str,
        space_table: pd.DataFrame,
        c_code: str,
        selection: pd.DataFrame,
        target_name: str | None = None,
        map_table: pd.DataFrame | None = None,
    ) -> CodingSchemesManager:
        # TODO: handle missing values. Options:
        # 1. A missingness-aware CodingScheme (not preferred, requires new class, new logic, new tests).
        # 2. Hard-code replacement here (preferred, just replace 'nan' with 'MISSING').
        source_scheme = CodingScheme.from_table(
            name=name, table=space_table, code_selection=selection, c_code=c_code, c_desc=c_code
        )
        manager = CodingSchemesManager().add_scheme(source_scheme)
        if map_table is not None and target_name is not None:
            names = self.aux.scoped_names
            target_scheme = CodingScheme.from_table(
                name=target_name,
                table=map_table,
                c_code=names.mapped.column_name(c_code),
                c_desc=names.mapped.column_name(c_code),
            )
            code_map = CodeMap.from_table(
                source_scheme,
                target_scheme,
                c_source_code=c_code,
                c_target_code=names.mapped.column_name(c_code),
                table=map_table,
            )
            manager = manager.add_scheme(target_scheme).add_map(code_map)
        return manager

    def make_gender_scheme(self, data_connection: Any) -> CodingSchemesManager:
        gender_space_table = self.tables.static.gender_space(data_connection)
        return self._make_demographic_scheme(
            name=self.scheme.gender,
            space_table=gender_space_table,
            c_code=COLUMN.gender,
            selection=self.aux.selections.gender,
        )

    def make_ethnicity_scheme(self, data_connection: Any) -> CodingSchemesManager:
        race_space_table = self.tables.static.ethnicity_space(data_connection)
        return self._make_demographic_scheme(
            name=self.scheme.ethnicity,
            space_table=race_space_table,
            c_code=COLUMN.race,
            selection=self.aux.selections.ethnicity,
            target_name=self.aux.scoped_names.mapped.ethnicity,
            map_table=self.aux.maps.ethnicity,
        )

    def make_obs_scheme(self) -> CodingSchemesManager:
        obs = self.tables.obs
        return obs.register_scheme(
            name=self.scheme.obs, space_table=obs.space(), attributes_selection=self.aux.selections.obs
        )

    def make_icu_inputs_scheme(self) -> CodingSchemesManager:
        c_code = str(COLUMN.code)
        c_desc = str(COLUMN.description)
        scheme = CodingSchemeWithUOM.from_table(
            name=self.scheme.icu_inputs,
            table=self.aux.icu_inputs_uom_normalization_table,
            c_code=c_code,
            c_desc=c_desc,
            c_universal_unit=str(COLUMN.derived_universal_unit),
            c_unit=str(COLUMN.amount_unit),
            c_normalization_factor=str(COLUMN.derived_unit_normalization_factor),
            code_selection=self.aux.selections.icu_inputs,
        )
        manager = CodingSchemesManager().add_scheme(scheme)
        c_aggregation = self.aux.icu_inputs_aggregation_column
        map_data = self.aux.maps.icu_inputs
        if map_data is not None and c_aggregation is not None:
            target_names = self.aux.scoped_names.mapped
            c_target_code = target_names.column_name(c_code)
            c_target_desc = target_names.column_name(c_desc)
            target_scheme = CodingScheme.from_table(
                name=target_names.icu_inputs, table=map_data, c_code=c_target_code, c_desc=c_target_desc
            )
            codemap = ReducedCodeMapN1.from_table(
                scheme,
                target_scheme,
                c_source_code=c_code,
                c_target_code=c_target_code,
                c_target_agg=c_aggregation,
                table=map_data,
            )
            manager = manager.add_scheme(target_scheme).add_map(codemap)

        return manager

    def make_icu_procedures_scheme(self, data_connection: Any) -> CodingSchemesManager:
        space_table = self.tables.icu_procedures.space(data_connection)
        source_scheme = CodingScheme.from_table(
            name=self.scheme.icu_procedures,
            table=space_table,
            code_selection=self.aux.selections.icu_procedures,
            c_code=str(COLUMN.code),
            c_desc=str(COLUMN.description),
        )
        manager = CodingSchemesManager().add_scheme(source_scheme)
        map_table = self.aux.maps.icu_procedures
        if map_table is not None:
            target_names = self.aux.scoped_names.mapped
            target_scheme = CodingScheme.from_table(
                name=target_names.icu_procedures,
                table=map_table,
                c_code=target_names.column_name(str(COLUMN.code)),
                c_desc=target_names.column_name(str(COLUMN.description)),
            )
            code_map = CodeMap.from_table(
                source_scheme,
                target_scheme,
                c_source_code=str(COLUMN.code),
                c_target_code=target_names.column_name(str(COLUMN.code)),
                table=map_table,
            )
            manager = manager.add_scheme(target_scheme).add_map(code_map)
        return manager

    def make_hosp_procedures_scheme(self, manager: CodingSchemesManager) -> CodingSchemesManager:
        table = self.tables.hosp_procedures
        name = self.scheme.hosp_procedures
        target_name = self.aux.scoped_names.mapped.hosp_procedures
        mapping = self.aux.maps.hosp_procedures
        selection = self.aux.selections.hosp_procedures
        c_target_code = self.aux.scoped_names.mapped.column_name(str(COLUMN.code))
        c_target_desc = self.aux.scoped_names.mapped.column_name(str(COLUMN.description))
        return table.setup_schemes(
            manager,
            name=name,
            component_schemes={"9": "icd9pcs", "10": "icd10pcs"},
            infer_maps=("icd9pcs", "icd10pcs", "pr_ccs", "pr_flat_ccs"),
            target_name=target_name,
            mapping=mapping,
            selection=selection,
            c_code=COLUMN.code,
            c_version=COLUMN.version,
            c_target_code=c_target_code,
            c_target_desc=c_target_desc,
        )

    def make_dx_discharge_scheme(self, manager: CodingSchemesManager) -> CodingSchemesManager:
        table = self.tables.dx_discharge
        name = self.scheme.dx_discharge
        target_name = self.aux.scoped_names.mapped.dx_discharge
        mapping = self.aux.maps.dx_discharge
        selection = self.aux.selections.dx_discharge
        c_target_code = self.aux.scoped_names.mapped.column_name(str(COLUMN.code))
        c_target_desc = self.aux.scoped_names.mapped.column_name(str(COLUMN.description))
        m = table.setup_schemes(
            manager,
            name=name,
            component_schemes={"9": "icd9cm", "10": "icd10cm"},
            infer_maps=(
                "icd9cm",
                "icd10cm",  # 'dx_ccs',
                "dx_flat_ccs",
            ),
            target_name=target_name,
            mapping=mapping,
            selection=selection,
            c_code=COLUMN.code,
            c_version=COLUMN.version,
            c_target_code=c_target_code,
            c_target_desc=c_target_desc,
        )
        # OVERRIDE the low-quality mapping icd10cm->dx_ccs with a chained map icd10cm->icd9cm->dx_ccs
        m = m.add_chained_map(name, "icd9cm", "dx_ccs", overwrite=True)
        return m

    def make_all_schemes(self, data_connection: Any) -> CodingSchemesManager:
        # make standard ones.
        manager = setup_standard_icd_ccs()
        if self.scheme.hosp_procedures is not None:
            manager += self.make_hosp_procedures_scheme(manager)
        if self.scheme.dx_discharge is not None:
            manager += self.make_dx_discharge_scheme(manager)
        if self.scheme.gender is not None:
            manager += self.make_gender_scheme(data_connection)
        if self.scheme.ethnicity is not None:
            manager += self.make_ethnicity_scheme(data_connection)
        if self.scheme.icu_inputs is not None:
            manager += self.make_icu_inputs_scheme()
        if self.scheme.icu_procedures is not None:
            manager += self.make_icu_procedures_scheme(data_connection)
        if self.scheme.obs is not None:
            manager += self.make_obs_scheme()
        return manager


class MIMICResourceExploratory(AbstractConfig):
    tables: DatasetTablesResources

    def __init__(self, tables: DatasetTablesResources):
        self.tables = tables

    def supported_gender(self, data_connection: Any) -> pd.DataFrame:
        return self.tables.static.gender_space(data_connection)

    def supported_ethnicity(self, data_connection: Any) -> pd.DataFrame:
        return self.tables.static.ethnicity_space(data_connection)

    def obs_stats(self, data_connection: Any) -> pd.DataFrame:
        return self.tables.obs.stats(data_connection)

    def supported_obs_variables(self, data_connection: None = None) -> pd.DataFrame:
        return self.tables.obs.space(data_connection)

    def supported_icu_procedures(self, data_connection: Any) -> pd.DataFrame:
        return self.tables.icu_procedures.space(data_connection)

    def supported_icu_inputs(self, data_connection: Any) -> pd.DataFrame:
        return self.tables.icu_inputs.space(data_connection)

    def supported_hosp_procedures(self, data_connection: Any) -> pd.DataFrame:
        return self.tables.hosp_procedures.space(data_connection)

    def supported_dx_discharge(self, data_connection: Any) -> pd.DataFrame:
        return self.tables.dx_discharge.space(data_connection)


class MIMICDatasetCompiler(AbstractConfig):
    tables: DatasetTablesResources
    scheme: DatasetSchemeConfig

    def __init__(self, tables: DatasetTablesResources, scheme: DatasetSchemeConfig):
        self.tables = tables
        self.scheme = scheme

    def load_static(self, data_connection: Any, admissions: pd.DataFrame) -> pd.DataFrame:
        return self.tables.static(data_connection, admissions=admissions)

    def load_admissions(self, data_connection: Any) -> pd.DataFrame:
        return self.tables.admissions(data_connection)

    def load_dx_discharge(self, data_connection: Any, schemes_manager: CodingSchemesManager) -> pd.DataFrame:
        table = self.tables.dx_discharge
        return table(data_connection, schemes_manager=schemes_manager, mixed_scheme_name=self.scheme.dx_discharge)

    def load_obs(self, data_connection: Any) -> pd.DataFrame:
        return self.tables.obs(data_connection)

    def load_icu_procedures(self, data_connection: Any) -> pd.DataFrame:
        return self.tables.icu_procedures(data_connection)

    def load_icu_inputs(self, data_connection: Any) -> pd.DataFrame:
        return self.tables.icu_inputs(data_connection)

    def load_hosp_procedures(self, data_connection: Any, schemes_manager: CodingSchemesManager) -> pd.DataFrame:
        table = self.tables.hosp_procedures
        return table(data_connection, schemes_manager=schemes_manager, mixed_scheme_name=self.scheme.hosp_procedures)

    def load_tables(self, data_connection: Any, schemes_manager: CodingSchemesManager) -> DatasetTables:
        S = self.scheme
        hosp_procedures = self.load_hosp_procedures(data_connection, schemes_manager) if S.hosp_procedures else None
        icu_procedures = self.load_icu_procedures(data_connection) if S.icu_procedures else None
        icu_inputs = self.load_icu_inputs(data_connection) if S.icu_inputs else None
        obs = self.load_obs(data_connection) if S.obs else None
        admissions = self.load_admissions(data_connection)
        static = self.load_static(data_connection, admissions=admissions)
        dx_discharge = self.load_dx_discharge(data_connection, schemes_manager)
        return DatasetTables(
            static=static,
            admissions=admissions,
            dx_discharge=dx_discharge,
            obs=obs,
            icu_procedures=icu_procedures,
            icu_inputs=icu_inputs,
            hosp_procedures=hosp_procedures,
        )


def load_scheme_manager(
    tables: DatasetTablesResources,
    scheme: DatasetSchemeConfig,
    aux: MIMICDatasetAuxiliaryResources,
    data_connection: Any,
) -> CodingSchemesManager:
    schemes_resources = MIMICSchemeResources(tables=tables, scheme=scheme, aux=aux)
    return schemes_resources.make_all_schemes(data_connection=data_connection)


def load_tables(
    tables: DatasetTablesResources,
    scheme: DatasetSchemeConfig,
    schemes_manager: CodingSchemesManager,
    data_connection: Any,
) -> DatasetTables:
    compiler = MIMICDatasetCompiler(tables=tables, scheme=scheme)
    return compiler.load_tables(data_connection=data_connection, schemes_manager=schemes_manager)


def load_mimic(
    config: DatasetConfig, tables: DatasetTablesResources, aux: MIMICDatasetAuxiliaryResources, data_connection: Any
) -> tuple[Dataset, CodingSchemesManager]:
    manager = load_scheme_manager(tables, config.scheme, aux, data_connection)
    tables = load_tables(tables, config.scheme, manager, data_connection)
    return Dataset(tables=tables, config=config), manager



from collections.abc import ItemsView, Iterator, KeysView, Mapping, ValuesView
from types import MappingProxyType
from typing import Self, TypeVar

import pandas as pd

from .base import AbstractWithDataframeEquivalent


K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")


class AbstractFrozenDict(AbstractWithDataframeEquivalent, Mapping[K, V]):
    data: MappingProxyType[K, V]

    def __init__(self, data: Mapping[K, V]):
        self.data = MappingProxyType(data)

    def __getitem__(self, key: K) -> V:
        return self.data[key]

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterator[K]:
        return iter(self.data)

    def __contains__(self, key: object) -> bool:
        return key in self.data

    def get(self, key: K, default: V | None = None) -> V | None:  # type: ignore
        return self.data.get(key, default)

    def items(self) -> ItemsView[K, V]:
        return self.data.items()

    def keys(self) -> KeysView[K]:
        return self.data.keys()

    def values(self) -> ValuesView[V]:
        return self.data.values()


class FrozenDict11(AbstractFrozenDict[str, V]):
    def to_dataframe(self) -> pd.DataFrame:
        return pd.DataFrame(
            list(self.data.values()), columns=pd.Series(["value"]), index=pd.Series(list(self.data.keys()))
        ).sort_index()

    @classmethod
    def from_dataframe(cls, dataframe: pd.DataFrame) -> Self:
        return cls(dataframe["value"].to_dict())


class FrozenDict1N(AbstractFrozenDict[str, frozenset[V]]):
    data: MappingProxyType[str, frozenset[V]]

    def __init__(self, data: Mapping[str, frozenset[V] | set[V]]):
        super().__init__({k: frozenset(v) for k, v in data.items()})

    def to_dataframe(self) -> pd.DataFrame:
        return (
            pd.DataFrame([(k, item) for k, v in self.data.items() for item in v], columns=pd.Series(["key", "value"]))
            .sort_values(["key", "value"])
            .reset_index(drop=True)
        )

    @classmethod
    def from_dataframe(cls, dataframe: pd.DataFrame) -> Self:
        return cls(dataframe.groupby("key")["value"].apply(frozenset).to_dict())


class FrozenDict1NM(AbstractFrozenDict[str, MappingProxyType[str, float]]):
    data: MappingProxyType[str, MappingProxyType[str, float]]

    def __init__(self, data: Mapping[str, Mapping[str, float]]):
        super().__init__({k: MappingProxyType(v) for k, v in data.items()})

    def to_dataframe(self) -> pd.DataFrame:
        return (
            pd.DataFrame(
                [(k1, k2, v) for k1, kv in self.data.items() for k2, v in kv.items()],
                columns=pd.Series(["k1", "k2", "v"]),
            )
            .sort_values(["k1", "k2", "v"])
            .reset_index(drop=True)
        )

    @classmethod
    def from_dataframe(cls, dataframe: pd.DataFrame) -> Self:
        return cls(
            {
                k1: MappingProxyType({k2: v["v"].iloc[0].item() for k2, v in df2.groupby("k2")})  # type: ignore
                for k1, df2 in dataframe.groupby("k1")
            }
        )



from typing import Literal


DxMultiCCSName = Literal["dx_ccs"]  # diagnostics
PrMultiCCSName = Literal["pr_ccs"]  # procedures
DxFlatCCSName = Literal["dx_flat_ccs"]  # diagnostic
PrFlatCCSName = Literal["pr_flat_ccs"]  # procedures

ICD9CMName = Literal["icd9cm"]  # diagnostic
ICD9PCSName = Literal["icd9pcs"]  # procedures
ICD10CMName = Literal["icd10cm"]  # diagnostics
ICD10PCSName = Literal["icd10pcs"]  # procedures



import random
from typing import cast

import numpy as np
import numpy.random as nr
import pandas as pd

import ehrax as rx
from ehrax import CodingSchemeWithUOM


UOM = ["m", "s", "g", "mg", "KG", "ml"]


def scheme(name: str, codes: list[str]) -> rx.CodingScheme:
    return rx.CodingScheme(name=name, codes=tuple(sorted(codes)), desc=rx.FrozenDict11(dict(zip(codes, codes))))


def scheme_with_uom(name: str, codes: list[str]) -> rx.CodingScheme:
    universal_unit = rx.FrozenDict11({c: random.choice(UOM) for c in codes})
    uom_normalization_factor = rx.FrozenDict1NM({c: {u.lower(): 1.0 for u in random.sample(UOM, k=3)} for c in codes})
    return rx.CodingSchemeWithUOM(
        name=name,
        codes=tuple(sorted(codes)),
        desc=rx.FrozenDict11(dict(zip(codes, codes))),
        uom_normalization_factor=uom_normalization_factor,
        universal_unit=universal_unit,
    )


def outcome_extractor(dx_scheme: rx.CodingScheme) -> rx.FilterOutcomeMapData:
    name = f"{dx_scheme.name}_outcome"
    k = max(3, len(dx_scheme) - 1)
    random.seed(0)
    excluded = tuple(random.sample(dx_scheme.codes, k=k))
    return rx.FilterOutcomeMapData(name=name, base_name=dx_scheme.name, exclude_codes=excluded)


def sample_codes(scheme: rx.CodingScheme, n: int) -> list[str]:
    codes = scheme.codes
    return random.choices(codes, k=n)


def _dx_codes(dx_scheme: rx.CodingScheme):
    v = nr.binomial(1, 0.5, size=len(dx_scheme)).astype(bool)
    return rx.CodesVector(vec=v, scheme=dx_scheme.name)


def inpatient_binary_input(n: int, p: int, los_hours: float):
    starttime = np.array(sorted(nr.choice(np.linspace(0, los_hours, max(n + 1, 1000)), replace=False, size=n)))
    endtime = starttime + nr.uniform(0, los_hours - starttime, size=(n,))
    code_index = nr.choice(p, size=n, replace=True)
    return rx.InpatientInput(starttime=starttime, endtime=endtime, code_index=code_index)


def inpatient_rated_input(n: int, p: int, los_hours: float):
    bin_input = inpatient_binary_input(n, p, los_hours)
    return rx.InpatientInput(
        starttime=bin_input.starttime,
        endtime=bin_input.endtime,
        code_index=bin_input.code_index,
        rate=nr.uniform(0, 1, size=(n,)),
    )


def _singular_codevec(scheme: rx.CodingScheme) -> rx.CodesVector:
    return scheme.codeset2vec({random.choice(scheme.codes)})


def _icu_inputs(icu_inputs_scheme: rx.CodingScheme, n_timestamps: int, los_hours: float):
    return inpatient_rated_input(n_timestamps, len(icu_inputs_scheme), los_hours=los_hours)


def _proc(scheme: rx.CodingScheme, n_timestamps: int, los_hours: float):
    return inpatient_binary_input(n_timestamps, len(scheme), los_hours=los_hours)


def demographic_vector_config() -> rx.DemographicVectorConfig:
    flags = random.choices([True, False], k=3)
    return rx.DemographicVectorConfig(*flags)


def date_of_birth() -> pd.Timestamp:
    return pd.to_datetime(pd.Timestamp("now") - pd.to_timedelta(nr.randint(0, 100 * 365), unit="D"))


def _static_info(ethnicity: rx.CodesVector, gender: rx.CodesVector) -> rx.StaticInfo:
    return rx.StaticInfo(ethnicity=ethnicity, gender=gender, date_of_birth=date_of_birth())


def _dx_codes_history(dx_codes: rx.CodesVector):
    v = nr.binomial(1, 0.5, size=len(dx_codes.vec)).astype(bool)
    return rx.CodesVector(vec=v + dx_codes.vec, scheme=dx_codes.scheme)


def _outcome(
    outcome_extractor_: rx.FilterOutcomeMapData,
    dataset_scheme_manager: rx.CodingSchemesManager,
    dx_codes: rx.CodesVector,
):
    source_codes = dataset_scheme_manager.scheme[dx_codes.scheme].vec2codeset(dx_codes.vec)
    extractor = dataset_scheme_manager.outcome[dx_codes.scheme, outcome_extractor_.name]
    return extractor(source_codes)


def sample_subjects_dataframe(
    n: int, ethnicity_scheme: rx.CodingScheme, gender_scheme: rx.CodingScheme
) -> pd.DataFrame:
    return pd.DataFrame(
        {
            str(rx.COLUMN.subject_id): list(str(i) for i in range(n)),
            str(rx.COLUMN.race): random.choices(ethnicity_scheme.codes, k=n),
            str(rx.COLUMN.gender): random.choices(gender_scheme.codes, k=n),
            str(rx.COLUMN.date_of_birth): pd.to_datetime(
                random.choices(pd.date_range(start="1/1/1900", end="1/1/2000", freq="D"), k=n)
            ),
        }
    )


def sample_admissions_dataframe(subjects_df: pd.DataFrame, n: int, max_stay_days: int) -> pd.DataFrame:
    c_subject = str(rx.COLUMN.subject_id)
    c_admission = str(rx.COLUMN.admission_id)
    c_admission_time = str(rx.COLUMN.start_time)
    c_discharge_time = str(rx.COLUMN.end_time)
    admit_dates = pd.to_datetime(random.choices(pd.date_range(start="1/1/2000", end="1/1/2020", freq="D"), k=n))
    los = [random.uniform(0.5, max_stay_days * 0.99) for _ in range(n)]
    disch_dates = admit_dates + pd.to_timedelta(los, unit="D")

    return pd.DataFrame(
        {
            c_subject: random.choices(subjects_df[c_subject], k=n),
            c_admission: list(str(i) for i in range(n)),
            c_admission_time: admit_dates,
            c_discharge_time: disch_dates,
        }
    )


def sample_dx_dataframe(admissions_df: pd.DataFrame, dx_scheme: rx.CodingScheme, n: int) -> pd.DataFrame:
    dx_codes = sample_codes(dx_scheme, n)
    return pd.DataFrame(
        {
            str(rx.COLUMN.admission_id): random.choices(admissions_df[str(rx.COLUMN.admission_id)], k=n),
            str(rx.COLUMN.code): dx_codes,
        }
    )


def _sample_proc_dataframe(admissions_df: pd.DataFrame, scheme: rx.CodingScheme, n: int) -> pd.DataFrame:
    c_admission = str(rx.COLUMN.admission_id)
    c_admittime = str(rx.COLUMN.start_time)
    c_dischtime = str(rx.COLUMN.end_time)
    c_code = str(rx.COLUMN.code)
    c_start = str(rx.COLUMN.start_time)
    c_end = str(rx.COLUMN.end_time)
    codes = sample_codes(scheme, n)
    df_in = pd.DataFrame(
        {
            c_admission: random.choices(admissions_df[c_admission], k=n),
            c_code: codes,
            c_start: pd.Timestamp(0),
            c_end: pd.Timestamp(0),
        }
    )
    df = pd.merge(df_in, admissions_df[[c_admission, c_admittime, c_dischtime]], on=c_admission, suffixes=(None, "_y"))
    c_admittime = f"{c_admittime}_y" if c_admittime in df_in.columns else c_admittime
    c_dischtime = f"{c_dischtime}_y" if c_dischtime in df_in.columns else c_dischtime

    df["los"] = (df[c_dischtime] - df[c_admittime]).dt.total_seconds() / 3600
    relative_start = nr.uniform(0, df["los"].values.tolist(), size=n)
    df[c_start] = df[c_admittime] + pd.to_timedelta(relative_start, unit="hours")
    relative_end = nr.uniform(low=relative_start, high=df["los"].values.tolist(), size=n)
    df[c_end] = df[c_admittime] + pd.to_timedelta(relative_end, unit="hours")
    assert df[c_start].between(df[c_admittime], df[c_dischtime]).all()
    assert df[c_end].between(df[c_admittime], df[c_dischtime]).all()
    return df[[c_admission, c_code, c_start, c_end]]


def sample_icu_inputs_dataframe(admissions_df: pd.DataFrame, icu_input_scheme: rx.CodingScheme, n: int) -> pd.DataFrame:
    df = _sample_proc_dataframe(admissions_df, icu_input_scheme, n)
    c_amount = str(rx.COLUMN.amount)
    c_unit = str(rx.COLUMN.amount_unit)
    c_code = str(rx.COLUMN.code)
    df[c_amount] = np.random.uniform(low=0, high=1000, size=n)
    scheme = cast(CodingSchemeWithUOM, SCHEMES["icu_inputs"])
    normalizer = scheme.uom_normalization_factor
    units_map = {c: list(d.keys()) for c, d in normalizer.items()}
    df[c_unit] = df[c_code].map(lambda c: random.choice(units_map[c]))
    return df


def sample_obs_dataframe(admissions_df: pd.DataFrame, obs_scheme: rx.CodingScheme, n: int) -> pd.DataFrame:
    c_admission = str(rx.COLUMN.admission_id)
    c_admittime = str(rx.COLUMN.start_time)
    c_dischtime = str(rx.COLUMN.end_time)
    c_obs = str(rx.COLUMN.code)
    c_time = str(rx.COLUMN.time)
    c_value = str(rx.COLUMN.measurement)

    codes = sample_codes(obs_scheme, n)
    df_in = pd.DataFrame(
        {c_admission: random.choices(admissions_df[c_admission], k=n), c_obs: codes, c_time: pd.Timestamp(0)}
    )
    df = pd.merge(df_in, admissions_df[[c_admission, c_admittime, c_dischtime]], on=c_admission, suffixes=(None, "_y"))
    c_admittime = f"{c_admittime}_y" if c_admittime in df_in.columns else c_admittime
    c_dischtime = f"{c_dischtime}_y" if c_dischtime in df_in.columns else c_dischtime

    df["los"] = (df[c_dischtime] - df[c_admittime]).dt.total_seconds() / 3600
    relative_time = nr.uniform(0, df["los"].values.tolist(), size=n)
    df[c_time] = df[c_admittime] + pd.to_timedelta(relative_time, unit="hours")

    assert isinstance(obs_scheme, rx.NumericScheme), "Only numeric schemes are supported"
    df["obs_type"] = df[c_obs].map(obs_scheme.type_hint.data)
    df.loc[df.obs_type == "N", c_value] = np.random.uniform(low=0, high=1000, size=(df.obs_type == "N").sum())
    df.loc[df.obs_type.isin(("C", "O")), c_value] = random.choices([0, 1, 2], k=df.obs_type.isin(("C", "O")).sum())
    df.loc[df.obs_type == "B", c_value] = random.choices([0, 1], k=(df.obs_type == "B").sum())

    return df[[c_admission, c_obs, c_time, c_value]]


def _dataset_tables(
    dataset_scheme_config: rx.DatasetSchemeConfig,
    dataset_scheme_manager: rx.CodingSchemesManager,
    freqs: tuple[int, ...],
    max_stay_days: int,
) -> rx.DatasetTables:
    n_subjects, n_admission_per_subject, n_per_admission = freqs
    assert dataset_scheme_config.ethnicity is not None
    assert dataset_scheme_config.gender is not None
    assert dataset_scheme_config.dx_discharge is not None
    assert dataset_scheme_config.icu_procedures is not None
    assert dataset_scheme_config.icu_inputs is not None
    assert dataset_scheme_config.obs is not None
    assert dataset_scheme_config.hosp_procedures is not None
    subjects_df = sample_subjects_dataframe(
        n_subjects,
        dataset_scheme_manager.scheme[dataset_scheme_config.ethnicity],
        dataset_scheme_manager.scheme[dataset_scheme_config.gender],
    )
    admissions_df = sample_admissions_dataframe(subjects_df, n_admission_per_subject * n_subjects, max_stay_days)
    dx_df = sample_dx_dataframe(
        admissions_df,
        dataset_scheme_manager.scheme[dataset_scheme_config.dx_discharge],
        n_per_admission * n_subjects * n_admission_per_subject,
    )

    obs_df = sample_obs_dataframe(
        admissions_df,
        dataset_scheme_manager.scheme[dataset_scheme_config.obs],
        n_per_admission * n_subjects * n_admission_per_subject,
    )

    icu_proc_df = _sample_proc_dataframe(
        admissions_df,
        dataset_scheme_manager.scheme[dataset_scheme_config.icu_procedures],
        n_per_admission * n_subjects * n_admission_per_subject,
    )

    hosp_proc_df = _sample_proc_dataframe(
        admissions_df,
        dataset_scheme_manager.scheme[dataset_scheme_config.hosp_procedures],
        n_per_admission * n_subjects * n_admission_per_subject,
    )

    icu_inputs_df = sample_icu_inputs_dataframe(
        admissions_df,
        dataset_scheme_manager.scheme[dataset_scheme_config.icu_inputs],
        n_per_admission * n_subjects * n_admission_per_subject,
    )

    return rx.DatasetTables(
        static=subjects_df,
        admissions=admissions_df,
        dx_discharge=dx_df,
        obs=obs_df,
        icu_procedures=icu_proc_df,
        hosp_procedures=hosp_proc_df,
        icu_inputs=icu_inputs_df,
    )


def make_targets_schemes_with_maps(
    n_scheme_targets: dict[str, int], source_schemes: dict[str, rx.CodingScheme]
) -> tuple[dict[str, rx.CodingScheme], dict[str, rx.CodeMap]]:
    def make_target_scheme_with_map(
        size: int, space: str, source_scheme: rx.CodingScheme
    ) -> tuple[str, rx.CodingScheme, rx.CodeMap]:
        assert size <= len(source_scheme)
        target_name = f"{source_scheme.name}_target"
        target_codes = tuple(f"{source_scheme}_target_{i}" for i in range(size))
        target_desc = rx.FrozenDict11(dict(zip(target_codes, target_codes)))
        target_scheme = rx.CodingScheme(name=target_name, codes=target_codes, desc=target_desc)
        mapp = {c: {t} for c, t in zip(source_scheme.codes, target_codes)}
        mapp |= {c: {random.choice(target_codes)} for c in source_scheme.codes[len(target_codes) :]}
        map_data = rx.FrozenDict1N({c: {random.choice(target_codes)} for c in source_scheme.codes})
        code_map = rx.CodeMap(source_name=source_scheme.name, target_name=target_name, data=map_data)
        return space, target_scheme, code_map

    space, schemes, maps = zip(
        *list(
            make_target_scheme_with_map(size, space, source_schemes[space]) for space, size in n_scheme_targets.items()
        )
    )
    return dict(zip(space, schemes)), dict(zip(space, maps))


SCHEMES: dict[str, rx.CodingScheme] = dict(
    ethnicity=scheme("ethnicity", ["E1", "E2", "E3"]),
    gender=scheme("genderrrr", ["M", "F"]),
    dx_discharge=scheme("dx1", ["Dx1", "Dx2", "Dx3", "Dx4", "Dx5", "Dx6", "Dx7", "Dx8", "Dx9", "Dx10"]),
    hosp_procedures=scheme("hosp_proc1", ["HP1", "HP2", "HP3", "HP4", "HP5", "HP6"]),
    icu_procedures=scheme("icu_proc2", ["ICU1", "ICU2", "ICU3", "ICU4", "ICU5", "ICU6"]),
    icu_inputs=scheme_with_uom("icu_inputs", ["ICUI1", "ICUI2", "ICUI3", "ICUI4", "ICUI5", "ICUI6"]),
    obs=rx.NumericScheme(
        name="observation11",
        codes=tuple(sorted(("O1", "O2", "O3", "O4", "O5"))),
        type_hint=rx.FrozenDict11(  # type: ignore
            dict(zip(("O1", "O2", "O3", "O4", "O5"), ("B", "C", "O", "N", "N")))
        ),
    ),
)

TARGET_SCHEMES, TARGET_SCHEMES_MAPS = make_targets_schemes_with_maps(
    n_scheme_targets={"dx_discharge": 5, "obs": 5, "icu_procedures": 3, "icu_inputs": 3, "hosp_procedures": 3},
    source_schemes=SCHEMES,
)

TARGET_SCHEMES_MAPS["icu_inputs"] = rx.ReducedCodeMapN1.from_data(
    SCHEMES["icu_inputs"].name,
    TARGET_SCHEMES["icu_inputs"].name,
    TARGET_SCHEMES_MAPS["icu_inputs"].data,
    rx.FrozenDict11({c: "w_sum" for c in TARGET_SCHEMES["icu_inputs"].codes}),
)
OUTCOME_DATA = outcome_extractor(TARGET_SCHEMES["dx_discharge"])
DATASET_SCHEME_MANAGER = rx.CodingSchemesManager(
    outcomes=(OUTCOME_DATA,),
    schemes=tuple(SCHEMES.values()) + tuple(TARGET_SCHEMES.values()),
    maps=tuple(TARGET_SCHEMES_MAPS.values()),
)

BINARY_OBSERVATION_CODE_INDEX = 0
CATEGORICAL_OBSERVATION_CODE_INDEX = 1
ORDINAL_OBSERVATION_CODE_INDEX = 2
NUMERIC_OBSERVATION_CODE_INDEX = 3

DATASET_SCHEME_CONF = rx.DatasetSchemeConfig(
    ethnicity=SCHEMES["ethnicity"].name,
    gender=SCHEMES["gender"].name,
    dx_discharge=SCHEMES["dx_discharge"].name,
    icu_procedures=SCHEMES["icu_procedures"].name,
    icu_inputs=SCHEMES["icu_inputs"].name,
    obs=SCHEMES["obs"].name,
    hosp_procedures=SCHEMES["hosp_procedures"].name,
)
DATASET_CONFIG = rx.DatasetConfig(scheme=DATASET_SCHEME_CONF)
TVXEHR_SCHEME_CONF = rx.TVxEHRSchemeConfig(
    ethnicity=SCHEMES["ethnicity"].name,
    gender=SCHEMES["gender"].name,
    dx_discharge=TARGET_SCHEMES["dx_discharge"].name,
    outcome=OUTCOME_DATA.name,
    icu_procedures=SCHEMES["icu_procedures"].name,
    icu_inputs=TARGET_SCHEMES["icu_inputs"].name,
    obs=SCHEMES["obs"].name,
    hosp_procedures=SCHEMES["hosp_procedures"].name,
)

TVXEHR_CONF = rx.TVxEHRConfig(scheme=TVXEHR_SCHEME_CONF, demographic=rx.DemographicVectorConfig())


def leading_observables_extractor(
    observation_scheme: rx.NumericScheme,
    leading_hours: tuple[float, ...] | list[float] = (1.0,),
    entry_neglect_window: float = 0.0,
    recovery_window: float = 0.0,
    minimum_acquisitions: int = 0,
    code_index: int = BINARY_OBSERVATION_CODE_INDEX,
) -> rx.LeadingObservableExtractor:
    config = rx.LeadingObservableExtractorConfig(
        observable_code=observation_scheme.codes[code_index],
        scheme=observation_scheme.name,
        entry_neglect_window=entry_neglect_window,
        recovery_window=recovery_window,
        minimum_acquisitions=minimum_acquisitions,
        leading_hours=leading_hours,
    )
    return rx.LeadingObservableExtractor(config=config, observable_scheme=observation_scheme)


def _inpatient_observables(observation_scheme: rx.CodingScheme, n_timestamps: int, los_hours: float):
    d = len(observation_scheme)
    timestamps_grid = np.linspace(0, los_hours, max(n_timestamps + 1, 1000), dtype=np.float64)
    t = np.array(sorted(nr.choice(timestamps_grid, replace=False, size=n_timestamps)))
    v = nr.randn(n_timestamps, d)
    mask = nr.binomial(1, 0.5, size=(n_timestamps, d)).astype(bool)
    return rx.InpatientObservables(t, v, mask)


def _inpatient_interventions(hosp_proc, icu_proc, icu_inputs):
    return rx.InpatientInterventions(hosp_proc, icu_proc, icu_inputs)


def _segmented_inpatient_interventions(
    inpatient_interventions: rx.InpatientInterventions,
    hosp_proc_scheme,
    icu_proc_scheme,
    icu_inputs_scheme,
    max_los_hours: float,
    maximum_padding: int = 1,
) -> rx.SegmentedInpatientInterventions:
    assert all(isinstance(s, rx.CodingScheme) for s in [hosp_proc_scheme, icu_proc_scheme, icu_inputs_scheme])
    return rx.SegmentedInpatientInterventions.from_interventions(
        inpatient_interventions,
        max_los_hours,
        hosp_procedures_size=len(hosp_proc_scheme),
        icu_procedures_size=len(icu_proc_scheme),
        icu_inputs_size=len(SCHEMES["icu_inputs"]),
        maximum_padding=maximum_padding,
    )


def _admission(
    admission_id: str,
    admission_date: pd.Timestamp,
    dx_codes: rx.CodesVector,
    dx_codes_history: rx.CodesVector,
    outcome: rx.CodesVector,
    observables: rx.InpatientObservables,
    interventions: rx.InpatientInterventions,
    leading_observable: rx.InpatientObservables,
    los_days: float,
) -> rx.Admission:
    discharge_date = pd.to_datetime(admission_date + pd.to_timedelta(los_days, unit="D") + pd.to_timedelta(1, unit="S"))

    return rx.Admission(
        admission_id=admission_id,
        admission_dates=rx.AdmissionDates(admission_date, discharge_date),
        dx_codes=dx_codes,
        dx_codes_history=dx_codes_history,
        outcome=outcome,
        observables=observables,
        interventions=interventions,
        leading_observable=leading_observable,
    )


def _admissions(
    n_admissions,
    dx_scheme: rx.CodingScheme,
    outcome_extractor_: rx.FilterOutcomeMap,
    observation_scheme: rx.NumericScheme,
    icu_inputs_scheme: rx.CodingScheme,
    icu_proc_scheme: rx.CodingScheme,
    hosp_proc_scheme: rx.CodingScheme,
    dataset_scheme_manager: rx.CodingSchemesManager,
    max_los_days: int,
    max_n_timestamps_obs: int,
    max_n_inputs: int,
) -> list[rx.Admission]:
    admissions = []
    multiplier = list(sorted(random.sample(range(10 * n_admissions), k=n_admissions)))
    admission_dates = [pd.to_datetime("now") + pd.to_timedelta(max_los_days * m, unit="D") for m in multiplier]
    for i, admission_date in enumerate(admission_dates):
        los_days = random.uniform(0, max_los_days)
        los_h = los_days * 24.0
        dx_codes = _dx_codes(dx_scheme)
        obs = _inpatient_observables(
            observation_scheme, n_timestamps=nr.randint(0, max_n_timestamps_obs), los_hours=los_h
        )
        lead = leading_observables_extractor(observation_scheme=observation_scheme)(obs)
        icu_proc = _proc(icu_proc_scheme, n_timestamps=nr.randint(0, max_n_inputs), los_hours=los_h)
        hosp_proc = _proc(hosp_proc_scheme, n_timestamps=nr.randint(0, max_n_inputs), los_hours=los_h)
        icu_inputs = _icu_inputs(icu_inputs_scheme, n_timestamps=nr.randint(0, max_n_inputs), los_hours=los_h)
        admissions.append(
            _admission(
                admission_id=f"test_{i}",
                admission_date=admission_date,
                dx_codes=dx_codes,
                dx_codes_history=_dx_codes_history(dx_codes),
                outcome=_outcome(outcome_extractor_, dataset_scheme_manager, dx_codes),
                observables=obs,
                los_days=los_days,
                interventions=_inpatient_interventions(hosp_proc=hosp_proc, icu_proc=icu_proc, icu_inputs=icu_inputs),
                leading_observable=lead,
            )
        )
    return admissions



from typing import cast, Literal

import pandas as pd

from ..base import AbstractConfig, AbstractVxData
from ..dataset import AdmissionsTableColumns, COLUMN, StaticTableColumns, TableColumns
from ..freezer import FrozenDict11
from .mimic import (
    CodedColumns,
    CodedTableResource,
    DatasetTablesResources,
    MixedICDTableResource,
    MixedICDTableResource_MIMICIII,
    MixedICDTableResource_MIMICIV,
    StaticTableResource,
    StaticTableResource_MIMICIII,
    StaticTableResource_MIMICIV,
    TableResource,
)


TableFileTitle = Literal["patients", "admissions", "diagnoses_icd", "d_icd_diagnoses"]


class InMemoryMIMICTableFiles(AbstractVxData):
    patients: pd.DataFrame
    admissions: pd.DataFrame
    diagnoses_icd: pd.DataFrame
    d_icd_diagnoses: pd.DataFrame

    def __init__(
        self,
        patients: pd.DataFrame,
        admissions: pd.DataFrame,
        diagnoses_icd: pd.DataFrame,
        d_icd_diagnoses: pd.DataFrame,
    ):
        self.patients = patients
        self.admissions = admissions
        self.diagnoses_icd = diagnoses_icd
        self.d_icd_diagnoses = d_icd_diagnoses

    @classmethod
    def from_path(
        cls,
        patients: str,
        admissions: str,
        diagnoses_icd: str,
        d_icd_diagnoses: str,
        usecols: dict[str, tuple[str, ...]] | None = None,
    ):
        if usecols is None:
            usecols = {}
        return InMemoryMIMICTableFiles(
            patients=pd.read_csv(patients, usecols=usecols.get("patients", None)),
            admissions=pd.read_csv(admissions, usecols=usecols.get("admissions", None)),
            diagnoses_icd=pd.read_csv(diagnoses_icd, usecols=usecols.get("diagnoses_icd", None)),
            d_icd_diagnoses=pd.read_csv(d_icd_diagnoses, usecols=usecols.get("d_icd_diagnoses", None)),
        )


class TableInterface(AbstractConfig):
    table_name: TableFileTitle
    column_map: FrozenDict11[str]

    def __init__(self, table_name: TableFileTitle, column_map: FrozenDict11[str]):
        self.table_name = table_name
        self.column_map = column_map

    def load_standard_columns_table(self, in_memory_tables: InMemoryMIMICTableFiles) -> pd.DataFrame:
        table = cast(pd.DataFrame, getattr(in_memory_tables, self.table_name))
        assert set(self.column_map.keys()).issubset(table.columns)
        return table.rename(columns=self.column_map)


class CodedTableInterface(TableInterface):
    table_name: TableFileTitle
    column_map: FrozenDict11[str]
    space_table_name: TableFileTitle
    space_column_map: FrozenDict11[str]

    def __init__(
        self,
        table_name: TableFileTitle,
        column_map: FrozenDict11[str],
        space_table_name: TableFileTitle,
        space_column_map: FrozenDict11[str],
    ):
        super().__init__(table_name=table_name, column_map=column_map)
        self.space_table_name = space_table_name
        self.space_column_map = space_column_map

    def load_space_table(self, in_memory_tables: InMemoryMIMICTableFiles) -> pd.DataFrame:
        table = getattr(in_memory_tables, self.space_table_name)
        assert set(self.space_column_map.keys()).issubset(table.columns)
        return table.rename(columns=self.space_column_map)[list(self.space_column_map.values())]


class StaticTableInterface(TableInterface):
    admissions_column_map: FrozenDict11[str]

    def __init__(self, static_column_map: FrozenDict11[str], admissions_column_map: FrozenDict11[str]):
        super().__init__(table_name="patients", column_map=static_column_map)
        self.admissions_column_map = admissions_column_map

    def load_standard_columns_table(self, in_memory_tables: InMemoryMIMICTableFiles) -> pd.DataFrame:
        assert set(self.column_map.keys()).issubset(in_memory_tables.patients)
        assert set(self.admissions_column_map.keys()).issubset(in_memory_tables.admissions)
        patients = in_memory_tables.patients.rename(columns=self.column_map)
        admissions = in_memory_tables.admissions.rename(columns=self.admissions_column_map)
        ethno_map = admissions.set_index(str(COLUMN.subject_id))[COLUMN.race].to_dict()
        patients[COLUMN.race] = patients[COLUMN.subject_id].map(ethno_map)
        return patients

    def load_gender_space_table(self, in_memory_tables: InMemoryMIMICTableFiles) -> pd.DataFrame:
        table = self.load_standard_columns_table(in_memory_tables)
        return table[[COLUMN.gender]].drop_duplicates()

    def load_ethnicity_space_table(self, in_memory_tables: InMemoryMIMICTableFiles) -> pd.DataFrame:
        table = self.load_standard_columns_table(in_memory_tables)
        return table[[COLUMN.race]].drop_duplicates()


class InMemoryTableResource(TableResource):
    interface: TableInterface

    def __init__(self, columns: TableColumns, table_name: TableFileTitle, column_map: FrozenDict11[str]):
        super().__init__(columns)
        self.interface = TableInterface(table_name=table_name, column_map=column_map)

    def load_standard_columns_table(self, in_memory_tables: InMemoryMIMICTableFiles, *args, **kwargs) -> pd.DataFrame:
        return self.interface.load_standard_columns_table(in_memory_tables)


class InMemoryCodedTableResource(CodedTableResource):
    interface: CodedTableInterface

    def __init__(
        self,
        columns: CodedColumns,
        table_name: TableFileTitle,
        column_map: FrozenDict11[str],
        space_table_name: TableFileTitle,
        space_column_map: FrozenDict11[str],
    ):
        super().__init__(columns)
        self.interface = CodedTableInterface(
            table_name=table_name,
            column_map=column_map,
            space_table_name=space_table_name,
            space_column_map=space_column_map,
        )

    def load_standard_columns_table(self, in_memory_tables: InMemoryMIMICTableFiles, *args, **kwargs) -> pd.DataFrame:
        return self.interface.load_standard_columns_table(in_memory_tables)

    def load_space_table(self, in_memory_tables: InMemoryMIMICTableFiles) -> pd.DataFrame:
        return self.interface.load_space_table(in_memory_tables)


class InMemoryStaticTableResource(AbstractConfig):  # mixin
    columns: StaticTableColumns
    interface: StaticTableInterface

    def __init__(self, static_column_map: FrozenDict11[str], admissions_column_map: FrozenDict11[str]):
        self.interface = StaticTableInterface(
            admissions_column_map=admissions_column_map, static_column_map=static_column_map
        )
        self.columns = StaticTableColumns()

    def load_standard_columns_table(self, in_memory_tables: InMemoryMIMICTableFiles, *args, **kwargs) -> pd.DataFrame:
        return self.interface.load_standard_columns_table(in_memory_tables)

    def load_gender_space_table(self, in_memory_tables: InMemoryMIMICTableFiles, *args, **kwargs) -> pd.DataFrame:
        return self.interface.load_gender_space_table(in_memory_tables)

    def load_ethnicity_space_table(self, in_memory_tables: InMemoryMIMICTableFiles, *args, **kwargs) -> pd.DataFrame:
        return self.interface.load_ethnicity_space_table(in_memory_tables)


class InMemoryStaticTableResource_MIMICIII(InMemoryStaticTableResource, StaticTableResource_MIMICIII):
    pass


class InMemoryStaticTableResource_MIMICIV(InMemoryStaticTableResource, StaticTableResource_MIMICIV):
    pass


class InMemoryMixedICDTableResource(AbstractConfig):  # mixin
    interface: CodedTableInterface

    def __init__(
        self,
        table_name: TableFileTitle,
        column_map: FrozenDict11[str],
        space_table_name: TableFileTitle,
        space_column_map: FrozenDict11[str],
    ):
        super().__init__()
        self.interface = CodedTableInterface(
            table_name=table_name,
            column_map=column_map,
            space_table_name=space_table_name,
            space_column_map=space_column_map,
        )

    def load_standard_columns_table(self, in_memory_tables: InMemoryMIMICTableFiles, *args, **kwargs) -> pd.DataFrame:
        return self.interface.load_standard_columns_table(in_memory_tables)

    def load_space_table(self, in_memory_tables: InMemoryMIMICTableFiles) -> pd.DataFrame:
        return self.interface.load_space_table(in_memory_tables)


class InMemoryMixedICDTableResource_MIMICIII(InMemoryMixedICDTableResource, MixedICDTableResource_MIMICIII):
    pass


class InMemoryMixedICDTableResource_MIMICIV(InMemoryMixedICDTableResource, MixedICDTableResource_MIMICIV):
    pass


class MIMICTablesResources(DatasetTablesResources):
    static: StaticTableResource
    admissions: TableResource
    dx_discharge: MixedICDTableResource
    obs: None
    hosp_procedures: None
    icu_procedures: None
    icu_inputs: None

    def __init__(self, static: StaticTableResource, admissions: TableResource, dx_discharge: MixedICDTableResource):
        super().__init__(
            static=static,
            admissions=admissions,
            dx_discharge=dx_discharge,
            obs=None,
            icu_procedures=None,
            icu_inputs=None,
            hosp_procedures=None,
        )


# The configurations below adapt to MIMIC-III v1.4
MIMICIII_STATIC_COLMAP = FrozenDict11(
    {
        "DOB": str(COLUMN.date_of_birth),
        "SUBJECT_ID": str(COLUMN.subject_id),
        "GENDER": str(COLUMN.gender),
    }
)
MIMICIII_ADMISSIONS_COLMAP = FrozenDict11(
    {
        "HADM_ID": str(COLUMN.admission_id),
        "SUBJECT_ID": str(COLUMN.subject_id),
        "ADMITTIME": str(COLUMN.start_time),
        "DISCHTIME": str(COLUMN.end_time),
        "ETHNICITY": str(COLUMN.race),
    }
)
MIMICIII_DIAGNOSES_ICD_COLMAP = FrozenDict11({"HADM_ID": str(COLUMN.admission_id), "ICD9_CODE": str(COLUMN.code)})
MIMICIII_D_ICD_DIAGNOSES_COLMAP = FrozenDict11({"ICD9_CODE": str(COLUMN.code), "LONG_TITLE": str(COLUMN.description)})

MIMICIII_STATIC_RESOURCES = InMemoryStaticTableResource_MIMICIII(MIMICIII_STATIC_COLMAP, MIMICIII_ADMISSIONS_COLMAP)

MIMICIII_ADMISSIONS_RESOURCES = InMemoryTableResource(
    AdmissionsTableColumns(), "admissions", MIMICIII_ADMISSIONS_COLMAP
)
MIMICIII_DX_DISCHARGE_RESOURCES = InMemoryMixedICDTableResource_MIMICIII(
    "diagnoses_icd", MIMICIII_DIAGNOSES_ICD_COLMAP, "d_icd_diagnoses", MIMICIII_D_ICD_DIAGNOSES_COLMAP
)
MIMICIII_TABLES_RESOURCES = MIMICTablesResources(
    static=MIMICIII_STATIC_RESOURCES,
    admissions=MIMICIII_ADMISSIONS_RESOURCES,
    dx_discharge=MIMICIII_DX_DISCHARGE_RESOURCES,
)

# The configurations below adapt to MIMIC-IV v3.1
MIMICIV_STATIC_COLMAP = FrozenDict11(
    {
        "subject_id": str(COLUMN.subject_id),
        "gender": str(COLUMN.gender),
        "anchor_age": str(COLUMN.anchor_age),
        "anchor_year": str(COLUMN.anchor_year),
    }
)

MIMICIV_ADMISSIONS_COLMAP = FrozenDict11(
    {
        "hadm_id": str(COLUMN.admission_id),
        "subject_id": str(COLUMN.subject_id),
        "admittime": str(COLUMN.start_time),
        "dischtime": str(COLUMN.end_time),
        "race": str(COLUMN.race),
    }
)

MIMICIV_DIAGNOSES_ICD_COLMAP = FrozenDict11(
    {"hadm_id": str(COLUMN.admission_id), "icd_code": str(COLUMN.code), "icd_version": str(COLUMN.version)}
)

MIMICIV_D_ICD_DIAGNOSES_COLMAP = FrozenDict11(
    {"icd_code": str(COLUMN.code), "icd_version": str(COLUMN.version), "long_title": str(COLUMN.description)}
)

MIMICIV_STATIC_RESOURCES = InMemoryStaticTableResource_MIMICIV(MIMICIV_STATIC_COLMAP, MIMICIV_ADMISSIONS_COLMAP)

MIMICIV_ADMISSIONS_RESOURCES = InMemoryTableResource(AdmissionsTableColumns(), "admissions", MIMICIV_ADMISSIONS_COLMAP)
MIMICIV_DX_DISCHARGE_RESOURCES = InMemoryMixedICDTableResource_MIMICIV(
    "diagnoses_icd", MIMICIV_DIAGNOSES_ICD_COLMAP, "d_icd_diagnoses", MIMICIV_D_ICD_DIAGNOSES_COLMAP
)
MIMICIV_TABLES_RESOURCES = MIMICTablesResources(
    static=MIMICIV_STATIC_RESOURCES,
    admissions=MIMICIV_ADMISSIONS_RESOURCES,
    dx_discharge=MIMICIV_DX_DISCHARGE_RESOURCES,
)



from collections import namedtuple
from typing import Any, cast

import ehrax as rx
import numpy as np
import pandas as pd
import pytest
from ehrax import CodingScheme, CodingSchemesManager, COLUMN, FrozenDict1N, FrozenDict11
from ehrax.example_datasets.mimic import (
    MIMICDatasetSchemeSuffixes,
    MixedICDTableResource,
    MixedICDTableResource_MIMICIII,
    ScopedSchemeNames,
    TableResource,
)
from ehrax.example_datasets.mimic_in_memory import MIMICIII_DX_DISCHARGE_RESOURCES
from ehrax.example_schemes.icd9 import ICD9CM
from ehrax.example_schemes.icd10 import ICD10CM
from ehrax.example_schemes.icd_ccs_integration import ICDSchemeSelection, setup_standard_icd_ccs
from ehrax.example_schemes.mixed_icd import MultiVersionScheme
from ehrax.utils import resources_path


@pytest.fixture(scope="module")
def standard_icd_manager() -> CodingSchemesManager:
    return setup_standard_icd_ccs(icd_selection=ICDSchemeSelection(icd9cm=True, icd10cm=True))


@pytest.fixture(scope="module")
def icd9cm(standard_icd_manager: CodingSchemesManager) -> CodingScheme:
    return standard_icd_manager.scheme["icd9cm"]


@pytest.fixture(scope="module")
def icd10cm(standard_icd_manager: CodingSchemesManager) -> CodingScheme:
    return standard_icd_manager.scheme["icd10cm"]


class TestTableResource:
    @pytest.mark.parametrize(
        "df, columns",
        [
            (pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}), ["a", "b"]),
            (pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}), ["a"]),
            (pd.DataFrame({"a": [], "b": []}), ["a"]),
            (pd.DataFrame({"a": ["1", "2"], "b": [4, 5]}), ["a"]),
        ],
    )
    def test_coerce_columns_to_str(self, df: pd.DataFrame, columns: list[str]):
        df2 = TableResource._coerce_columns_to_str(df, columns)
        assert all(pd.api.types.is_string_dtype(df2.dtypes[c]) for c in columns)

    def test_preprocess_pipeline(self):
        add1 = lambda x: x + 1
        assert TableResource.apply_pipeline((add1, add1, add1), 2) == 5


ICD9_KEY = "9"
ICD10_KEY = "10"
N_CODES_PER_SCHEME = 10
MIXED_SCHEME_NAME = "tesTooOO"


class TestMixedICDTableResource:
    @pytest.fixture(scope="class")
    def icd_version_schemes(self) -> FrozenDict11:
        return FrozenDict11({ICD9_KEY: "icd9cm", ICD10_KEY: "icd10cm"})

    @pytest.fixture(scope="class")
    def supported_space(
        self, standard_icd_manager: CodingSchemesManager, icd_version_schemes: FrozenDict11
    ) -> pd.DataFrame:
        def codes_df(version: str, codes: tuple[str, ...], desc: tuple[str, ...]) -> pd.DataFrame:
            return pd.DataFrame(
                {
                    str(rx.COLUMN.code): codes,
                    str(rx.COLUMN.version): [version] * len(codes),
                    str(rx.COLUMN.description): desc,
                }
            )

        S = standard_icd_manager.scheme
        codes = {name: S[name].codes[:N_CODES_PER_SCHEME] for name in icd_version_schemes.values()}
        return pd.concat(
            [
                codes_df(v, codes[name], tuple(map(S[name].desc.get, codes[name])))
                for v, name in icd_version_schemes.items()
            ]
        )

    @pytest.fixture(scope="class")
    def registered_schemes(
        self,
        standard_icd_manager: CodingSchemesManager,
        icd_version_schemes: FrozenDict11,
        supported_space: pd.DataFrame,
    ) -> CodingSchemesManager:
        scheme = MixedICDTableResource.register_scheme(
            MIXED_SCHEME_NAME,
            {k: standard_icd_manager.scheme[v] for k, v in icd_version_schemes.items()},
            supported_space,
        )
        manager = standard_icd_manager.add_scheme(scheme)
        for name in icd_version_schemes.values():
            manager = scheme.register_infer_map(manager, name)
        return manager

    @pytest.fixture(scope="class")
    def mixed_scheme(self, registered_schemes: CodingSchemesManager) -> MultiVersionScheme:
        return cast(MultiVersionScheme, registered_schemes.scheme[MIXED_SCHEME_NAME])

    @pytest.fixture(scope="class")
    def mixed_code_space(
        self, registered_schemes: CodingSchemesManager, mixed_scheme: MultiVersionScheme, supported_space: pd.DataFrame
    ) -> pd.DataFrame:
        return mixed_scheme.mixed_code_format_table(registered_schemes, supported_space)

    def test_mixed_code_space(
        self,
        supported_space: pd.DataFrame,
        mixed_code_space: pd.DataFrame,
        mixed_scheme: MultiVersionScheme,
        icd9cm: CodingScheme,
        icd10cm: CodingScheme,
    ):
        assert supported_space.shape == mixed_code_space.shape
        codes = supported_space[COLUMN.code]
        assert (codes.isin(icd9cm.codes).astype(int) + codes.isin(icd10cm.codes).astype(int) == 1).all()
        assert mixed_code_space[COLUMN.code].isin(mixed_scheme.codes).all()

    def test_mixed_scheme_properties(self, mixed_scheme: MultiVersionScheme):
        assert isinstance(mixed_scheme, MultiVersionScheme)
        assert len(mixed_scheme) == N_CODES_PER_SCHEME * 2

    def test_mixed_schemes_maps(
        self,
        mixed_scheme: MultiVersionScheme,
        registered_schemes: CodingSchemesManager,
        icd9cm: CodingScheme,
        icd10cm: CodingScheme,
    ):
        assert (mixed_scheme.name, icd9cm.name) in registered_schemes.map
        assert (mixed_scheme.name, icd10cm.name) in registered_schemes.map

    def test_mixed_codes_reversal(
        self,
        mixed_scheme: MultiVersionScheme,
        registered_schemes: CodingSchemesManager,
        mixed_code_space: pd.DataFrame,
        supported_space: pd.DataFrame,
    ):
        mixed_codes = mixed_code_space[COLUMN.code]
        mixed_as_icd9 = mixed_codes.map(registered_schemes.map[MIXED_SCHEME_NAME, "icd9cm"].data).to_numpy()
        mixed_as_icd10 = mixed_codes.map(registered_schemes.map[MIXED_SCHEME_NAME, "icd10cm"].data).to_numpy()
        versions = supported_space[COLUMN.version].to_numpy()

        df = supported_space.copy()
        codes = np.where(versions == ICD9_KEY, mixed_as_icd9, mixed_as_icd10)
        assert all(map(lambda c: len(c) == 1, codes))
        codes = list(map(lambda c: next(iter(c)), codes))
        df[COLUMN.code] = codes
        assert df.equals(supported_space)


class TestScopedNames:
    @pytest.fixture(scope="class")
    def suffixes(self) -> MIMICDatasetSchemeSuffixes:
        return MIMICDatasetSchemeSuffixes(
            gender="bin_gender",
            ethnicity="ethnicity",
            dx_discharge="dx_mixed_icd",
            obs="obs",
            icu_inputs="icu_inputs",
            icu_procedures=None,
            hosp_procedures="pr_mixed_icd",
        )

    @pytest.fixture(scope="class", params=[".", ":", "xxx", "**", "-", "_", "!", "%", "/", "no justice no peace"])
    def prefix(self, request) -> str:
        return request.param

    @pytest.fixture(scope="class", params=["_", ":", "***", "-"])
    def name_separator(self, request) -> str:
        return request.param

    @pytest.fixture(scope="class", params=[(), ("why",)])
    def global_infix(self, request) -> tuple[str, ...]:
        return request.param

    @pytest.fixture(scope="class")
    def variable_scoped_name(
        self, suffixes: MIMICDatasetSchemeSuffixes, global_infix: tuple[str, ...], name_separator: str, prefix: str
    ) -> ScopedSchemeNames:
        return ScopedSchemeNames(
            suffixes=suffixes, global_infix=global_infix, name_separator=name_separator, name_prefix=prefix
        )

    @pytest.fixture(scope="class")
    def constant_scoped_name(self, suffixes) -> ScopedSchemeNames:
        return ScopedSchemeNames(suffixes=suffixes, global_infix=(), name_separator="_", name_prefix="SSS")

    def test_scoped_names0(self, constant_scoped_name: ScopedSchemeNames, suffixes: MIMICDatasetSchemeSuffixes):
        assert constant_scoped_name.gender == "SSS_bin_gender"
        assert constant_scoped_name.ethnicity == "SSS_ethnicity"
        assert constant_scoped_name.dx_discharge == "SSS_dx_mixed_icd"
        assert constant_scoped_name.icu_procedures is None

    def test_scoped_names(
        self,
        variable_scoped_name: ScopedSchemeNames,
        suffixes: MIMICDatasetSchemeSuffixes,
        global_infix: tuple[str, ...],
        prefix: str,
        name_separator: str,
    ):
        infix = f"{name_separator.join(global_infix)}{name_separator}" if len(global_infix) > 0 else ""
        assert variable_scoped_name.gender == f"{prefix}{name_separator}{infix}{suffixes.gender}"

    def test_scoped_names_mapped(
        self,
        variable_scoped_name: ScopedSchemeNames,
        suffixes: MIMICDatasetSchemeSuffixes,
        global_infix: tuple[str, ...],
        prefix: str,
        name_separator: str,
    ):
        infix = f"{name_separator.join(global_infix)}{name_separator}" if len(global_infix) > 0 else ""
        mapped = variable_scoped_name.mapped
        assert mapped.gender == f"{prefix}{name_separator}{infix}mapped{name_separator}{suffixes.gender}"


MIXED_SCHEME_NAME = "XYZ"


class TestMIMIC3Diagnoses:
    @pytest.fixture(scope="class")
    def diagnoses_icd(self) -> pd.DataFrame:
        return pd.read_csv(resources_path("testing", "DIAGNOSES_ICD.csv"))

    @pytest.fixture(scope="class")
    def d_icd_diagnoses(self, diagnoses_icd: pd.DataFrame) -> pd.DataFrame:
        codes = diagnoses_icd[diagnoses_icd["ICD9_CODE"].notnull()]["ICD9_CODE"].unique()
        df = pd.DataFrame({"ICD9_CODE": codes})
        df["ROW_ID"] = range(len(df))
        df["SHORT_TITLE"] = "bla"
        df["LONG_TITLE"] = "blabla"
        return df

    @pytest.fixture(scope="class")
    def in_memory_data(self, diagnoses_icd: pd.DataFrame, d_icd_diagnoses: pd.DataFrame) -> Any:
        C = namedtuple("InMemoryData", ["diagnoses_icd", "d_icd_diagnoses"])
        return C(diagnoses_icd, d_icd_diagnoses)

    @pytest.fixture(scope="class")
    def diagnoses_icd_std(self, in_memory_data: Any, table_resource: MixedICDTableResource_MIMICIII) -> pd.DataFrame:
        return table_resource.load_standard_columns_table(in_memory_data)

    @pytest.fixture(scope="class")
    def d_icd_diagnoses_std(self, in_memory_data: Any, table_resource: MixedICDTableResource) -> pd.DataFrame:
        return table_resource.load_space_table(in_memory_data)

    @pytest.fixture(scope="class")
    def table_resource(self) -> MixedICDTableResource_MIMICIII:
        return MIMICIII_DX_DISCHARGE_RESOURCES

    @pytest.fixture(scope="class")
    def light_icd9_scheme(self, d_icd_diagnoses: pd.DataFrame) -> CodingScheme:
        codes = d_icd_diagnoses["ICD9_CODE"]
        codes = codes[codes.notnull()].tolist()
        codes = list(map(ICD9CM.format, codes))
        return ICD9CM(
            name="light_icd9", codes=tuple(sorted(codes)), ch2pt=FrozenDict1N({c: {codes[-1]} for c in codes[:-1]})
        )

    @pytest.fixture(scope="class")
    def light_icd10_scheme(self, d_icd_diagnoses: pd.DataFrame) -> CodingScheme:
        return ICD10CM(name="light_icd10", codes=(), ch2pt=FrozenDict1N({}))

    @pytest.fixture(scope="class")
    def init_manager(self, light_icd9_scheme: CodingScheme, light_icd10_scheme: CodingScheme) -> CodingSchemesManager:
        return CodingSchemesManager().add_scheme(light_icd9_scheme).add_scheme(light_icd10_scheme)

    @pytest.fixture(scope="class")
    def updated_manager(
        self, init_manager: CodingSchemesManager, in_memory_data: Any, table_resource: MixedICDTableResource
    ) -> CodingSchemesManager:
        scheme = table_resource.register_scheme(
            name=MIXED_SCHEME_NAME,
            component_schemes={"9": init_manager.scheme["light_icd9"]},
            selection=table_resource.space(in_memory_data),
        )
        manager = init_manager.add_scheme(scheme)
        manager = scheme.register_infer_map(manager, "light_icd9")
        return manager

    @pytest.fixture(scope="class")
    def after_coerce_id2str(
        self, diagnoses_icd_std: pd.DataFrame, table_resource: MixedICDTableResource_MIMICIII
    ) -> pd.DataFrame:
        return table_resource._coerce_id_to_str(diagnoses_icd_std)

    @pytest.fixture(scope="class")
    def after_nullcode_filter(self, diagnoses_icd_std: pd.DataFrame) -> pd.DataFrame:
        return TableResource._filter_null_codes(diagnoses_icd_std)

    @pytest.fixture(scope="class")
    def after_add_version_column(self, diagnoses_icd_std: pd.DataFrame):
        return MixedICDTableResource_MIMICIII._add_version_column_if_not_exists(diagnoses_icd_std)

    @pytest.fixture(scope="class")
    def mixed_format(
        self,
        diagnoses_icd_std: pd.DataFrame,
        updated_manager: CodingSchemesManager,
        table_resource: MixedICDTableResource_MIMICIII,
    ) -> CodingSchemesManager:
        pipeline = (
            MixedICDTableResource_MIMICIII._filter_null_codes,
            table_resource._coerce_code_to_str,
            MixedICDTableResource_MIMICIII._strip_icd_codes,
            MixedICDTableResource_MIMICIII._add_version_column_if_not_exists,
            MixedICDTableResource_MIMICIII._coerce_version_to_str,
            MixedICDTableResource_MIMICIII._mixed_code_format(MIXED_SCHEME_NAME, updated_manager),
        )
        return TableResource.apply_pipeline(pipeline, diagnoses_icd_std)

    def test_std_icd_diagnoses(self, diagnoses_icd_std: pd.DataFrame, diagnoses_icd: pd.DataFrame):
        # no change in content.
        assert diagnoses_icd.shape == diagnoses_icd_std.shape
        df_a = diagnoses_icd_std.iloc[:]
        df_b = diagnoses_icd.iloc[:]
        df_a.columns = df_b.columns
        assert df_a.equals(df_b)
        # columns updated to COLUMN type.
        assert {COLUMN.admission_id, COLUMN.code}.issubset(diagnoses_icd_std.columns)
        assert len({COLUMN.admission_id, COLUMN.code}.intersection(diagnoses_icd.columns)) == 0

    def test_std_d_icd_diagnoses(self, d_icd_diagnoses: pd.DataFrame, d_icd_diagnoses_std: pd.DataFrame):
        # no content changed.
        assert len(d_icd_diagnoses) == len(d_icd_diagnoses_std)
        df_a = d_icd_diagnoses_std.loc[:, [COLUMN.code, COLUMN.description]]
        df_b = d_icd_diagnoses.loc[:, ["ICD9_CODE", "LONG_TITLE"]]
        df_a.columns = df_b.columns
        assert df_a.equals(df_b)

    def test_coerce_id2str(self, diagnoses_icd_std: pd.DataFrame, after_coerce_id2str: pd.DataFrame):
        # does not change column names
        assert all(after_coerce_id2str.columns == after_coerce_id2str.columns)
        # content type changes
        assert not diagnoses_icd_std.equals(after_coerce_id2str)
        assert any(diagnoses_icd_std.dtypes != after_coerce_id2str.dtypes)
        assert all(type(i) is int for i in diagnoses_icd_std[COLUMN.admission_id])
        assert all(type(i) is str for i in after_coerce_id2str[COLUMN.admission_id])

    def test_filter_null_codes(self, diagnoses_icd_std: pd.DataFrame, after_nullcode_filter: pd.DataFrame):
        # no change in column names.
        assert all(after_nullcode_filter.columns == after_nullcode_filter.columns)
        # no change in dtypes
        assert all(after_nullcode_filter.dtypes == after_nullcode_filter.dtypes)
        # one row contained an empty code.
        assert len(after_nullcode_filter) + 1 == len(diagnoses_icd_std)

    def test_add_version_column(self, diagnoses_icd_std: pd.DataFrame, after_add_version_column: pd.DataFrame):
        assert set(diagnoses_icd_std.columns) == set(after_add_version_column.columns) - {COLUMN.version}
        assert all(after_add_version_column[COLUMN.version] == "9")
        assert diagnoses_icd_std.equals(after_add_version_column[diagnoses_icd_std.columns])

    def test_coerce_version2str(
        self,
        diagnoses_icd_std: pd.DataFrame,
    ):
        df = diagnoses_icd_std.iloc[:]
        df[COLUMN.version] = 9
        df2 = MixedICDTableResource_MIMICIII._coerce_version_to_str(df)
        assert all(type(v) is str for v in df2[COLUMN.version])

    def test_mixed_format(
        self, mixed_format: pd.DataFrame, diagnoses_icd_std: pd.DataFrame, updated_manager: CodingSchemesManager
    ):
        pipeline = (
            MixedICDTableResource_MIMICIII._filter_null_codes,
            lambda df: MixedICDTableResource_MIMICIII._coerce_columns_to_str(df, (COLUMN.code,)),
            MixedICDTableResource_MIMICIII._strip_icd_codes,
            MixedICDTableResource_MIMICIII._add_version_column_if_not_exists,
            MixedICDTableResource_MIMICIII._coerce_version_to_str,
        )
        pure_format = TableResource.apply_pipeline(pipeline, diagnoses_icd_std)
        assert pure_format.shape == mixed_format.shape
        assert all(pure_format.columns == mixed_format.columns)
        assert all(pure_format.dtypes == mixed_format.dtypes)
        rest_cols = list(set(pure_format.columns) - {COLUMN.code})
        assert pure_format[rest_cols].reset_index(drop=True).equals(mixed_format[rest_cols].reset_index(drop=True))
        mix_map = updated_manager.map[(MIXED_SCHEME_NAME, "light_icd9")]
        pure_codes = tuple(map(ICD9CM.format, pure_format[COLUMN.code].tolist()))
        mixed_codes = mixed_format[COLUMN.code].tolist()
        (unmixed_codes,) = zip(*list(map(mix_map.data.get, mixed_codes)))
        assert unmixed_codes == pure_codes

    def test_pipeline_without_null_code_filtration(
        self, updated_manager: CodingSchemesManager, diagnoses_icd_std: pd.DataFrame
    ):
        bad_pipeline = (
            lambda df: MixedICDTableResource_MIMICIII._coerce_columns_to_str(df, (COLUMN.code,)),
            MixedICDTableResource_MIMICIII._strip_icd_codes,
            MixedICDTableResource_MIMICIII._add_version_column_if_not_exists,
            MixedICDTableResource_MIMICIII._coerce_version_to_str,
            MixedICDTableResource_MIMICIII._mixed_code_format(MIXED_SCHEME_NAME, updated_manager),
        )
        good_pipeline = (MixedICDTableResource_MIMICIII._filter_null_codes,) + bad_pipeline
        with pytest.raises(Exception):
            _ = TableResource.apply_pipeline(bad_pipeline, diagnoses_icd_std)
        df = TableResource.apply_pipeline(good_pipeline, diagnoses_icd_std)
        assert isinstance(df, pd.DataFrame)



ADMISSION_TABLES_MAX_STAY_DAYS = 356
ADMISSION_CONCEPT_MAX_STAY_DAYS = 10
OBS_MAX_N_TIMESTAMPS = 1000
INTERVENTIONS_MAX_N_ITEMS = 50
ADMISSION_CONCEPT_MAX_STAY_HOURS = ADMISSION_CONCEPT_MAX_STAY_DAYS * 24



import json
import os
from collections.abc import Iterable, Mapping
from functools import cached_property
from types import MappingProxyType
from typing import Self

import networkx as nx
import pandas as pd

from ..coding_scheme import FrozenDict1N, FrozenDict11, HierarchicalScheme
from ..utils import tqdm_constructor


TERMS_DICT = {
    "T-00000": "SNOMED RT+CTV3",
    "T-01000": "body structure",
    "T-01100": "morphologic abnormality",
    "T-01200": "cell structure",
    "T-01210": "cell",
    "T-02000": "finding",
    "T-02100": "disorder",
    "T-03000": "environment / location",
    "T-03100": "environment",
    "T-03200": "geographic location",
    "T-04000": "event",
    "T-05000": "observable entity",
    "T-06000": "organism",
    "T-07000": "product",
    "T-07100": "medicinal product",
    "T-07110": "medicinal product form",
    "T-07111": "clinical drug",
    "T-08000": "physical force",
    "T-09000": "physical object",
    "T-10000": "procedure",
    "T-10100": "regime/therapy",
    "T-11000": "qualifier value",
    "T-11100": "administration method",
    "T-11200": "disposition",
    "T-11300": "intended site",
    "T-11800": "supplier",
    "T-11900": "product name",
    "T-11400": "release characteristic",
    "T-11500": "transformation",
    "T-11020": "basic dose form",
    "T-11030": "dose form",
    "T-11600": "role",
    "T-11700": "state of matter",
    "T-11040": "unit of presentation",
    "T-12000": "record artifact",
    "T-13000": "situation",
    "T-14000": "metadata",
    "T-14100": "core metadata concept",
    "T-14200": "foundation metadata concept",
    "T-14300": "linkage concept",
    "T-14310": "attribute",
    "T-14320": "link assertion",
    "T-14400": "namespace concept",
    "T-14500": "OWL metadata concept",
    "T-15000": "social concept",
    "T-15100": "life style",
    "T-15010": "racial group",
    "T-15020": "ethnic group",
    "T-15200": "occupation",
    "T-15300": "person",
    "T-15400": "religion/philosophy",
    "T-16000": "special concept",
    "T-16100": "inactive concept",
    "T-16200": "navigational concept",
    "T-17000": "specimen",
    "T-18000": "staging scale",
    "T-18100": "assessment scale",
    "T-18200": "tumor staging",
    "T-19000": "substance",
}

DESCRIPTION_RELATION = "900000000000003001"
SYNONYM_RELATION = "900000000000013009"
IS_A_RELATION = "116680003"

OTHER_RELATIONS = {
    "scale_type": "S-370132008",  # Scale type
    "property": "S-370130000",  # Property
    "inheres_in": "S-704319004",  # Inheres-in
    "inherent_location": "S-718497002",  # Inherent location
    "characterizes": "S-704321009",  # Characterises
    "direct_site": "S-704327008",  # Direct-Site
    "process_output": "S-704324001",  # Process-output
    "units": "S-246514001",  # Units
}


class SNOMEDCTGBMonolith:
    @staticmethod
    def link_with_desc(terms: pd.DataFrame, desc: pd.DataFrame, active: bool = True) -> pd.DataFrame:
        """
        terms is the basic table of concepts. The table consists of the following columns:

        |Field              |Data type  |Purpose    |Mutable|Part of Primary Key|
        |-------------------|-----------|-----------|-------|-------------------|
        |id                 |SCTID      |Uniquely Idenfies the concept|NO|YES (Full/Snapshot)|
        |effectiveTime      |Time       |Specifies the inclusive date at which the component version's state became the
                then current valid state of the component.|YES|YES (Full)<br>Optional (Snapshot)|
        |active             |Boolean    |Specifies whether the concept was active or inactive from the nominal release
                date specified by the effectiveTime.|YES|NO|
        |moduleId           |SCTID      |Identifies the concept version's module. Set to a descendant of
                900000000000443000(Module) within the metadata hierarchy.|YES|NO|
        |definitionStatusId |SCTID      |Specifies if the concept version is primitive or defined. Set to a descendant
                of 900000000000444006(Definition status)in the metadata hierarchy.|YES|NO|

        desc is the table of descriptions linked to each concept. The table consists of the following columns:

        |Field              |Data type  |Purpose    |Mutable|Part of Primary Key|
        |-------------------|-----------|-----------|-------|-------------------|
        |id                 |SCTID      |Uniquely identifies the description.|NO|YES (Full/Snapshot)|
        |effectiveTime      |Time       |Specifies the inclusive date at which the component version's state became the
                then current valid state of the component|YES|YES (Full)<br>Optional |Snapshot||
        |active             |Boolean    |Specifies whether the state of the description was active or inactive from the
                nominal release date specified by the effectiveTime.|YES|NO|
        |moduleId           |SCTID      |Identifies the description version's module. Set to a child of
                900000000000443000|Module| within the metadata hierarchy.|YES|NO|
        |conceptId          |SCTID      |Identifies the concept to which this description applies. Set to the identifier
                of a concept in the 138875005 |SNOMED CT Concept| hierarchy within the Concept. Note that a specific
                version of a description is not directly bound to a specific version of the concept to which it applies.
                Which version of a description applies to a concept depends on its effectiveTime and the point in time
                at which it is accessed.|NO|NO|
        |languageCode       |String     |Specifies the language of the description text using the two character
                ISO-639-1 code. Note that this specifies a language level only, not a dialect or country code.|NO|NO|
        |typeId             |SCTID      |Identifies whether the description is fully specified name a synonym or other
                description type. This field is set to a child of 900000000000446008|Description type| in the
                Metadata hierarchy.|NO|NO|
        |term               |String     |The description version's text value, represented in UTF-8 encoding.|YES|NO|
        |caseSignificanceId |SCTID      |Identifies the concept enumeration value that represents the case
                significance of this description version. For example, the term may be completely case sensitive, case
                insensitive or initial letter case insensitive. This field will be set to a child of
                900000000000447004|Case significance| within the metadata hierarchy.|YES|NO|


        Taken from: https://confluence.ihtsdotools.org/display/DOCRELFMT
        """

        def link_all_desc(terms_df: pd.DataFrame, desc_df: pd.DataFrame) -> pd.DataFrame:
            _ = pd.merge(terms_df, desc_df, left_on=["id"], right_on=["conceptId"], how="inner")
            with_primary_desc = _[_["typeId"] == DESCRIPTION_RELATION]
            with_primary_desc = with_primary_desc.drop_duplicates(["id_x"], keep="first")
            with_synonym_desc = _[_["typeId"] == SYNONYM_RELATION]
            return pd.concat([with_primary_desc, with_synonym_desc])

        with_desc = pd.merge(
            terms, desc[desc["typeId"] == DESCRIPTION_RELATION], left_on=["id"], right_on=["conceptId"], how="inner"
        )
        with_desc = with_desc.drop_duplicates(["id_x"], keep="first")
        assert not active or len(with_desc) == len(terms)
        with_desc = with_desc.assign(tui=with_desc["term"].str.extract(r"\((\w+\s?.?\s?\w+.?\w+.?\w+.?)\)$"))
        with_all_desc = link_all_desc(terms, desc)
        # Check if there are the same amount of active concepts
        assert not active or len(with_all_desc[with_all_desc["typeId"] == DESCRIPTION_RELATION]) == len(terms)
        snomed_cdb_df = pd.merge(with_all_desc, with_desc, left_on=["id_x"], right_on=["conceptId"], how="inner")
        # clean up the merge and rename the columns to fit the medcat Concept database criteria
        rename_cols = {"id_x_x": "cui", "term_x": "str", "typeId_x": "tty", "tui": "sty"}
        snomed_cdb_df = snomed_cdb_df.loc[:, list(rename_cols.keys())].rename(columns=rename_cols)

        # Add tui codes
        dict2 = {v: k for k, v in TERMS_DICT.items()}
        return snomed_cdb_df.assign(
            onto="SNOMED-CT",
            tty=snomed_cdb_df["tty"].str.replace(DESCRIPTION_RELATION, "1").str.replace(SYNONYM_RELATION, "0"),
            cui="S-" + snomed_cdb_df["cui"].astype(str),
            tui=snomed_cdb_df["sty"].map(dict2),
        )

    @staticmethod
    def parse_file(fname, first_row_header=True, columns=None) -> pd.DataFrame:
        with open(fname, encoding="utf-8") as f:
            entities = [[n.strip() for n in line.split("\t")] for line in f]
            return pd.DataFrame(entities[1:], columns=entities[0] if first_row_header else columns)

    @classmethod
    def load(
        cls, monolith_dir: str, write_disk: bool = True
    ) -> tuple[pd.DataFrame, pd.DataFrame, dict[str, list[str]], Mapping[str, Mapping[str, str]]]:
        """
        # TODO: cleanup + factorise.
        To understand the SNOMED-CT organisation/philosophy: https://confluence.ihtsdotools.org/display/DOCRELFMT
        ## SNOMED CT Design

            ### SNOMED CT Components
            SNOMED CT is a clinical terminology containing concepts with unique meanings and formal logic based
                definitions organised into hierarchies.
            For further information please see: https://confluence.ihtsdotools.org/display/DOCSTART/4.+SNOMED+CT+Basics

            SNOMED CT content is represented into 3 main types of components:
            - __Concepts__ representing clinical meanings that are organised into hierarchies.
            - __Descriptions__ which link appropriate human-readable terms to concepts
            - __Relationships__ which link each concept to other related concepts

        """

        def filename(l: list[str], prefix: str) -> str:
            match = [f for f in l if f.lower().startswith(prefix)]
            assert len(match) == 1
            return match[0]

        term_dir = f"{monolith_dir}/Snapshot/Terminology"
        term_dir_files = os.listdir(term_dir)
        concept_file = os.path.join(term_dir, filename(term_dir_files, "sct2_concept_"))
        description_file = os.path.join(term_dir, filename(term_dir_files, "sct2_description_"))
        terms = cls.parse_file(concept_file)
        desc = cls.parse_file(description_file)

        active_terms = terms[terms.active == "1"]  # active concepts are represented with 1
        inactive_terms = terms[terms.active != "1"]
        active_descs = desc[desc.active == "1"]
        inactive_descs = desc[desc.active != "1"]

        snomed_cdb_active_df = cls.link_with_desc(active_terms, active_descs, active=True)
        snomed_cdb_inactive_df = cls.link_with_desc(inactive_terms, inactive_descs, active=False)

        ###################
        ### Relations
        ###################
        # |Field                |Data type  |Purpose|Mutable|Part of Primary Key|
        # |---------------------|-----------|-------|-------|-----|
        # |id                   |SCTID      |Uniquely identifies the relationship.|NO|YES(Full/Snapshot)|
        # |effectiveTime        |Time       |Specifies the inclusive date at which the component version's state
        #           became the then current valid state of the component.|YES|YES(Full) Optional(Snapshot)|
        # |active               |Boolean    |Specifies whether the state of the relationship was active or inactive
        #       from the nominal release date specified by the effectiveTime field.|YES|NO|
        # |moduleId             |SCTID      |Identifies the relationship version's module. Set to a child of
        #       900000000000443000|Module| within the metadata hierarchy.|YES|NO|
        # |sourceId             |SCTID      |Identifies the source concept of the relationship version. That is the
        #       concept defined by this relationship. Set to the identifier of a concept.|NO|NO|
        # |destinationId |SCTID      |Identifies the concept that is the destination of the relationship version.
        #       <br>That is the concept representing the value of the attribute represented by the typeId column.
        #       <br>Set to the identifier of a concept.<br>Note that the values that can be applied to particular
        #       attributes are formally defined by the SNOMED CT Machine Readable Concept Model.|NO|NO|
        # |relationshipGroup    |Integer    |Groups together relationship versions that are part of a
        #       logically associated relationshipGroup. All active Relationship records with the same relationship Group
        #       number and sourceId are grouped in this way.|YES|NO|
        # |typeId               |SCTID      |Identifies the concept that represent the defining attribute (or
        #       relationship type) represented by this relationship version.<br><br>That is the concept representing
        #       the value of the attribute represented by the typeId column. <br><br>Set to the identifier of a concept.
        #       The concept identified must be either 116680003|Is a| or a subtype of
        #       410662002|Concept model attribute|. The concepts that can be used as in the typeId column are formally
        #       defined as follows:<br>116680003|is a| OR < 410662002|concept model attribute|<br><br>__Note__ that the
        #       attributes that can be applied to particular concepts are formally defined by the SNOMED CT Machine
        #       Readable Concept Model.|NO|NO|
        # |characteristicTypeId |SCTID      |A concept enumeration value that identifies the characteristic type of the
        #       relationship version (i.e. whether the relationship version is defining, qualifying, etc.) This field is
        #       set to a descendant of 900000000000449001|Characteristic type|in the metadata hierarchy.|YES|NO|
        # |modifierId           |SCTID      |A concept enumeration value that identifies the type of Description
        #       Logic(DL) restriction (some, all, etc.). Set to a child of 900000000000450001|Modifier| in the metadata
        #       hierarchy.<br> __Note__ Currently the only value used in this column is 900000000000451002|Some| and
        #       thus in practical terms this column can be ignored.|YES|NO|
        relations_file = os.path.join(term_dir, filename(term_dir_files, "sct2_relationship_"))
        rel = cls.parse_file(relations_file)
        is_a = rel.loc[(rel.active == "1") & (rel.typeId == IS_A_RELATION), ["sourceId", "destinationId"]].astype(str)
        is_a = "S-" + is_a
        ch2pt = is_a.groupby("sourceId")["destinationId"].apply(list).to_dict()

        if write_disk:
            # Write the clinical terms to csv
            snomed_cdb_active_df.to_csv("snomed_cdb_active.csv.gz", compression="gzip")
            snomed_cdb_inactive_df.to_csv("snomed_cdb_inactive.csv.gz", compression="gzip")

            # Write to 'isa' relationships to file
            with open("isa_active_rela_ch2pt.json", "w") as outfile:
                json.dump(dict(ch2pt), outfile)

        # Other relations.
        other_rel = rel.loc[rel.active == "1", ["sourceId", "destinationId", "typeId"]]
        other_rel = "S-" + other_rel
        other_rel = other_rel.loc[other_rel["typeId"].isin(OTHER_RELATIONS.values()), :]
        relation_label = {v: k for k, v in OTHER_RELATIONS.items()}
        other_relations = {
            relation_label[relation_type_id]: relations.set_index("sourceId")["destinationId"].to_dict()
            for relation_type_id, relations in other_rel.groupby("typeId")
        }

        return snomed_cdb_active_df, snomed_cdb_inactive_df, ch2pt, other_relations

    @classmethod
    def process_refset(cls, filename: str) -> pd.DataFrame:
        df = cls.parse_file(filename)
        df = df[df.active == "1"]
        df = df.rename(columns={"referencedComponentId": "member", "refsetId": "refset"})
        df = df[["member", "refset"]]
        for c in df.columns:
            df.loc[:, c] = df.loc[:, c].str.strip()
        return "S-" + df


class SNOMEDCT(HierarchicalScheme):
    cdb_df: pd.DataFrame
    cdb_inactive_df: pd.DataFrame
    active_terms: frozenset[str]
    other_relations: MappingProxyType[str, FrozenDict11[str]] | None

    def __init__(
        self,
        name: str,
        codes: tuple[str, ...],
        desc: FrozenDict11,
        cdb_df: pd.DataFrame,
        cdb_inactive_df: pd.DataFrame,
        active_terms: set[str],
        ch2pt: FrozenDict1N,
        other_relations: Mapping[str, Mapping[str, str]] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(name=name, codes=codes, desc=desc, ch2pt=ch2pt, **kwargs)
        self.cdb_df = cdb_df
        self.cdb_inactive_df = cdb_inactive_df
        self.active_terms = frozenset(active_terms)
        if other_relations is None:
            other_relations = MappingProxyType({k: MappingProxyType({}) for k in OTHER_RELATIONS.keys()})

        self.other_relations = other_relations

    def to_desc(self, m: Mapping[str, str]) -> Mapping[str, str]:
        return {k: self.desc.get(v) for k, v in m.items()}

    @cached_property
    def code_inheres_in(self) -> Mapping[str, str]:
        return self.to_desc(self.other_relations["inheres_in"])

    @cached_property
    def code_property(self) -> Mapping[str, str]:
        return self.to_desc(self.other_relations["property"])

    @cached_property
    def code_inherent_location(self) -> Mapping[str, str]:
        return self.to_desc(self.other_relations["inherent_location"])

    @cached_property
    def code_characterizes(self) -> Mapping[str, str]:
        return self.to_desc(self.other_relations["characterizes"])

    @cached_property
    def code_direct_site(self) -> Mapping[str, str]:
        return self.to_desc(self.other_relations["direct_site"])

    @cached_property
    def code_units(self) -> Mapping[str, str]:
        return self.to_desc(self.other_relations["units"])

    @classmethod
    def from_tables(
        cls,
        name: str,
        cdb_active: pd.DataFrame,
        cdb_inactive: pd.DataFrame,
        ch2pt: dict[str, list[str]],
        other_relations: Mapping[str, Mapping[str, str]],
    ) -> Self:
        cdb_df, active_terms, active_desc = cls.cdb_table(cdb_active)
        cdb_inactive_df, inactive_terms, inactive_desc = cls.cdb_table(cdb_inactive)
        # the active replaces inactive for any overlap
        desc = inactive_desc["name"].to_dict() | active_desc["name"].to_dict()
        return cls(
            name=name,
            codes=tuple(sorted(active_terms | inactive_terms)),
            desc=FrozenDict11(desc),
            cdb_df=cdb_df,
            cdb_inactive_df=cdb_inactive_df,
            active_terms=active_terms,
            ch2pt=FrozenDict1N({k: frozenset(v) for k, v in ch2pt.items()}),
            other_relations=MappingProxyType({k: MappingProxyType(v) for k, v in other_relations.items()}),
        )

    @classmethod
    def cdb_table(cls, cdb: pd.DataFrame) -> tuple[pd.DataFrame, set[str], pd.DataFrame]:
        terms = set(cdb.cui.unique())
        desc_table = cdb[cdb.tty == "1"].reset_index(drop=True)
        desc_table = desc_table.groupby("cui").agg(name=("str", lambda x: x.values[0]))
        return cdb, terms, desc_table

    @classmethod
    def from_files(cls, name: str, cdb_active_path: str, cdb_inactive_path: str, ch2pt_json_path: str) -> Self:
        cdb_active_table = pd.read_csv(cdb_active_path, index_col=0)
        cdb_inactive_table = pd.read_csv(cdb_inactive_path, index_col=0)
        with open(ch2pt_json_path) as json_file:
            return cls.from_tables(name, cdb_active_table, cdb_inactive_table, json.load(json_file))

    @classmethod
    def from_gb_monolith_dir(cls, name: str, gb_monolith_dir: str) -> Self:
        cdb_active, cdb_inactive, ch2pt, other_relations = SNOMEDCTGBMonolith.load(gb_monolith_dir, write_disk=False)
        return cls.from_tables(name, cdb_active, cdb_inactive, ch2pt, other_relations=other_relations)

    def to_networkx(
        self,
        codes: tuple[str, ...] = None,
        discard_set: set[str] | None = None,
        node_attrs: dict[str, dict[str, str]] | None = None,
    ) -> nx.DiGraph:
        """
        Generate a networkx.DiGraph (Directed Graph) from a table of SNOMED-CT codes.

        Args:
            codes (tuple[str, ...]): The table of codes, must have a column `core_code` for the SNOMED-CT codes.
            discard_set (Optional[set[str]]): A set of codes, which, if provided, they are excluded from
                the Graph object.
            node_attrs: A dictionary of node attributes, which, if provided, used to annotate nodes with additional
                information, such as the frequency of the corresponding SNOMED-CT code in a particular dataset.
        """

        if codes is None:
            codes = set(self.codes) & set(self.ch2pt.keys())

        def parents_traversal(x) -> frozenset[str]:
            ch2pt_edges = set()

            def parents_traversal_(node):
                for pt in self.ch2pt.get(node, ()):
                    ch2pt_edges.add((node, pt))
                    parents_traversal_(pt)

            parents_traversal_(x)
            return frozenset(ch2pt_edges)

        if discard_set:
            ch2pt_edges = [parents_traversal(c) for c in tqdm_constructor(c for c in codes if c not in discard_set)]
        else:
            ch2pt_edges = [parents_traversal(c) for c in tqdm_constructor(codes)]

        dag = nx.DiGraph()

        for ch, pt in frozenset().union(*ch2pt_edges):
            dag.add_edge(ch, pt)

        if node_attrs is not None:
            for node in tqdm_constructor(dag.nodes):
                for attr_name, attr_dict in node_attrs.items():
                    dag.nodes[node][attr_name] = attr_dict.get(node, "")
        return dag

    def as_dataframe(self, codes: Iterable[str] | None = None) -> pd.DataFrame:
        """
        Returns the scheme as a Pandas DataFrame.
        The DataFrame contains the following columns:
            - code: the code string
            - desc: the code description
        """
        if codes is None:
            codes = self.codes
        index = codes
        return pd.DataFrame(
            {
                "code": codes,
                "desc": list(map(self.desc.get, codes)),
                "inheres_in": self.code_inheres_in,
                "property": self.code_property,
                "units": self.code_units,
                "characetrizes": self.code_characterizes,
                "direct_site": self.code_direct_site,
                "inherent_location": self.code_inherent_location,
            },
            index=index,
        )



import gzip
import xml.etree.ElementTree as ET
from collections import defaultdict
from typing import Any

import pandas as pd

from ..coding_scheme import FrozenDict11
from ..freezer import FrozenDict1N
from ..utils import resources_path
from .icd import ICDHierarchicalScheme, ICDScheme


class ICD10PCS(ICDScheme):
    @staticmethod
    def format(code: str) -> str:
        # No decimal point in ICD10-PCS
        return code


class ICD10CM(ICDHierarchicalScheme):
    """
    NOTE: for prediction targets, remember to exclude the following chapters:
        - 'chapter:19': 'Injury, poisoning and certain \
            other consequences of external causes (S00-T88)',
        - 'chapter:20': 'External causes of morbidity (V00-Y99)',
        - 'chapter:21': 'Factors influencing health status and \
            contact with health services (Z00-Z99)',
        - 'chapter:22': 'Codes for special purposes (U00-U85)'
    """

    @staticmethod
    def format(code: str) -> str:
        code = code.upper()  # ICD10 is case insensitive
        if "." in code:
            # logging.debug(f'Code {code} already is in decimal format')
            return code
        if len(code) > 3:
            return code[:3] + "." + code[3:]
        else:
            return code


class ICD10CMFactory:
    FILE_TABLE_TITLE: str = "icd-10-cm-tabular-2025.xml.gz"
    FILE_LIST_TITLE: str = "icd10cm-codes-2025.txt.gz"
    FILE_LEGACY_MAP_TITLE: str = "icd-10-cm-conversion-table-FY2025.csv.gz"

    C_LIST_CODE: str = "C"
    C_LIST_DESCRIPTION: str = "D"

    C_LEGACY_CURRENT: str = "C"
    C_LEGACY_RETIRED: str = "R"
    C_LEGACY_EFFECTIVE: str = "E"

    @classmethod
    def expand_ranges1(cls, first: pd.Series, second: pd.Series) -> pd.Series:
        # with no period.
        pass

    @classmethod
    def expand_ranges2(cls, first: pd.Series, second: pd.Series) -> pd.Series:
        # with period.
        pass

    @classmethod
    def expand_ranges(cls, r: pd.Series):
        r = r.str.split("-")
        first = r.map(lambda x: x[0])
        second = r.map(lambda x: x[1])
        # either both sides contain a dot or both do not.
        assert all(first.str.contains(".") == second.str.contains("."))
        if any(first.str.contains(".")):
            return cls.expand_ranges2(first, second)
        return cls.expand_ranges1(first, second)

    @classmethod
    def load_legacy_table(cls) -> pd.DataFrame:
        df = pd.read_csv(resources_path("ICD", cls.FILE_LEGACY_MAP_TITLE), dtype=str, sep="\t", skiprows=1)
        df.columns = [cls.C_LEGACY_CURRENT, cls.C_LEGACY_EFFECTIVE, cls.C_LEGACY_RETIRED]
        for ci in (0, 1):
            df.iloc[:, ci] = df.iloc[:, ci].str.strip()
        df = df.loc[(df.loc[:, cls.C_LEGACY_CURRENT] != "None") & (df.loc[:, cls.C_LEGACY_RETIRED].notnull()), :]
        return df[df.iloc[:, 0].notnull()].reset_index(drop=True)

    @classmethod
    def parse_legacy_map_file(cls) -> dict[str, str]:
        pass

    @classmethod
    def load_codes_list_file(cls) -> pd.DataFrame:
        lines = []
        with gzip.open(resources_path("ICD", cls.FILE_LIST_TITLE), "rb") as f:
            for line in f.readlines():
                code = ICD10CM.format(line[:7].strip().decode("utf-8"))
                desc = line[7:].strip().decode("utf-8")
                lines.append((code, desc))
        return pd.DataFrame(lines, columns=[cls.C_LIST_CODE, cls.C_LIST_DESCRIPTION])

    @classmethod
    def augment_hierarchy(cls, icd10cm_tabular_traversal: dict[str, Any]) -> dict[str, Any]:
        codes_list = cls.load_codes_list_file()
        codes = icd10cm_tabular_traversal["codes"]
        desc = icd10cm_tabular_traversal["desc"]
        ch2pt = icd10cm_tabular_traversal["ch2pt"]

        # update description.
        desc.update(codes_list.set_index("C")["D"].to_dict())

        def update_hierarchy(c, pt):
            ch2pt[c] = frozenset({pt})

        codes_list = codes_list[~codes_list.C.isin(codes)]
        # what remains undetected are full codes (8 characters: 7 stem + 1 period).
        assert (codes_list.C.map(len) == 8).all()
        # the code 'ABC.DEFG' is one level higher than 'ABC.DEF'
        # the code 'AB' is only four levels higher than 'ABC.DEF'. (period is just a place holder).
        for height in (1, 2, 3, 5):  #
            potential_parents = codes_list.C.map(lambda c: c[:-height])
            detected_codes = potential_parents.isin(codes)
            for ch, pt in zip(codes_list.loc[detected_codes, "C"], potential_parents):
                update_hierarchy(ch, pt)
            codes_list = codes_list.loc[~detected_codes, :]
        assert len(codes_list) == 0, "All should be populated."
        return {
            "codes": tuple(sorted(desc.keys())),
            "desc": FrozenDict11(desc),
            "ch2pt": FrozenDict1N(ch2pt),
        }

    @classmethod
    def traverse_icd10_xml(cls) -> dict[str, Any]:
        # https://www.cdc.gov/nchs/icd/Comprehensive-listing-of-ICD-10-CM-Files.htm
        with gzip.open(resources_path("ICD", cls.FILE_TABLE_TITLE), "r") as f:
            tree = ET.parse(f)

        code_level = {}
        root = tree.getroot()
        ch2pt = defaultdict(list)
        root_node = f"root:{root.tag}"
        desc = {root_node: "root"}
        chapters = [ch for ch in root if ch.tag == "chapter"]

        def _traverse_diag_dfs(parent_name, dx_element):
            dx_name = next(e for e in dx_element if e.tag == "name").text
            dx_desc = next(e for e in dx_element if e.tag == "desc").text
            desc[dx_name] = dx_desc
            ch2pt[dx_name].append(parent_name)
            children = tuple(dx for dx in dx_element if dx.tag == "diag")
            code_level[dx_name] = "node" if len(children) > 0 else "leaf"
            for ch in children:
                _traverse_diag_dfs(dx_name, ch)

        for chapter in chapters:
            ch_name = next(e for e in chapter if e.tag == "name").text
            ch_desc = next(e for e in chapter if e.tag == "desc").text
            code_level[ch_name] = "chapter"
            ch2pt[ch_name].append(root_node)
            desc[ch_name] = ch_desc

            sections = [sec for sec in chapter if sec.tag == "section"]
            for section in sections:
                sec_name = section.attrib["id"]
                sec_desc = next(e for e in section if e.tag == "desc").text
                code_level[sec_name] = "section"
                ch2pt[sec_name].append(ch_name)
                desc[sec_name] = sec_desc

                for dx in (dx for dx in section if dx.tag == "diag"):
                    _traverse_diag_dfs(sec_name, dx)

        # valid (complete codes) and invalid icd10 codes (e.g. chapters, sections, and internal nodes).
        return {
            "codes": tuple(sorted(desc.keys())),
            "desc": desc,
            "ch2pt": {ch: frozenset(pts) for ch, pts in ch2pt.items()},
        }

    @classmethod
    def create_scheme(cls) -> ICD10CM:  # expose
        return ICD10CM(name="icd10cm", **cls.augment_hierarchy(cls.traverse_icd10_xml()))


class ICD10PCSFactory:
    FILE_TITLE: str = "icd10pcs_codes_2024.txt.gz"

    @classmethod
    def distill_icd10_xml(cls) -> dict[str, Any]:
        with gzip.open(resources_path("ICD", cls.FILE_TITLE), "rt") as f:
            desc = {code: desc for code, desc in map(lambda line: line.strip().split(" ", 1), f.readlines())}
            return {"codes": tuple(sorted(desc)), "desc": FrozenDict11(desc)}

    @classmethod
    def create_scheme(cls) -> ICD10PCS:  # expose
        return ICD10PCS(name="icd10pcs", **cls.distill_icd10_xml())
